from underthesea.utils.vietnamese_ipa import Syllable


def viet2ipa(text: str, *args, **kwargs):
    """Generate ipa of the syllable

    Vietnamese syllabic structure (Anh & Trang 2022)

    syllable = onset + rhyme + tone

    rhyme = medial + nuclear vowel + (coda)

    Args:
        text (str): represents syllable
        dialect (str): Either the `string` `"north"` or `"south"`. Default: `north`
        eight (boolean): If true, use eight tone format, else use six tone format. Default: `False`
        tone (str): Either the `string` `"ipa"` or `"number"`. Default: `number`

    Returns:
        A `string`. Represents ipa of the syllable

    Examples:

        >>> # -*- coding: utf-8 -*-
        >>> from underthesea.pipeline.ipa import viet2ipa
        >>> viet2ipa("trồng")
        tɕoŋ³²
    """
    try:
        syllable = Syllable(text)
        ipa = syllable.generate_ipa(*args, **kwargs)
        return ipa
    except Exception:
        return ""



import json
import pickle

import haiku as hk
import jax
import jax.numpy as jnp

from .config import FLAGS
from .model import Generator


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def mel2wave(
    mel,
    config_file="assets/hifigan/config.json",
    ckpt_file=FLAGS.ckpt_dir / "hk_hifi.pickle",
):
    with open(config_file) as f:
        data = f.read()
    json_config = json.loads(data)
    h = AttrDict(json_config)

    @hk.transform_with_state
    def forward(x):
        net = Generator(h)
        return net(x)

    rng = next(hk.PRNGSequence(42))

    with open(ckpt_file, "rb") as f:
        params = pickle.load(f)
    aux = {}
    wav, aux = forward.apply(params, aux, rng, mel)
    wav = jnp.squeeze(wav)
    audio = jax.device_get(wav)
    return audio



# flake8: noqa

import random
from pathlib import Path

import numpy as np
from scipy.io import wavfile

from .config import FLAGS, AcousticInput, DurationInput


def load_phonemes_set():
    S = FLAGS.special_phonemes + FLAGS._normal_phonemes
    return S


def pad_seq(s, maxlen, value=0):
    assert maxlen >= len(s)
    return tuple(s) + (value,) * (maxlen - len(s))


def is_in_word(phone, word):
    def time_in_word(time, word):
        return (word.minTime - 1e-3) < time and (word.maxTime + 1e-3) > time

    return time_in_word(phone.minTime, word) and time_in_word(phone.maxTime, word)


def load_textgrid(fn: Path):
    """load textgrid file"""
    import textgrid
    tg = textgrid.TextGrid.fromFile(str(fn.resolve()))
    data = []
    words = list(tg[0])
    widx = 0
    assert tg[1][0].minTime == 0, "The first phoneme has to start at time 0"
    for p in tg[1]:
        if not p in words[widx]:
            widx = widx + 1
            if len(words[widx - 1].mark) > 0:
                data.append((FLAGS.special_phonemes[FLAGS.word_end_index], 0.0))
            if widx >= len(words):
                break
            assert p in words[widx], "mismatched word vs phoneme"
        mark = p.mark.strip().lower()
        if len(mark) == 0:
            mark = "sil"
        data.append((mark, p.duration()))
    return data


def textgrid_data_loader(data_dir: Path, seq_len: int, batch_size: int, mode: str):
    """load all textgrid files in the directory"""
    tg_files = sorted(data_dir.glob("*.TextGrid"))
    random.Random(42).shuffle(tg_files)
    L = len(tg_files) * 95 // 100
    assert mode in ["train", "val"]
    phonemes = load_phonemes_set()
    if mode == "train":
        tg_files = tg_files[:L]
    if mode == "val":
        tg_files = tg_files[L:]

    data = []
    for fn in tg_files:
        ps, ds = zip(*load_textgrid(fn))
        ps = [phonemes.index(p) for p in ps]
        l = len(ps)
        ps = pad_seq(ps, seq_len, 0)
        ds = pad_seq(ds, seq_len, 0)
        data.append((ps, ds, l))

    batch = []
    while True:
        random.shuffle(data)
        for e in data:
            batch.append(e)
            if len(batch) == batch_size:
                ps, ds, lengths = zip(*batch)
                ps = np.array(ps, dtype=np.int32)
                ds = np.array(ds, dtype=np.float32)
                lengths = np.array(lengths, dtype=np.int32)
                yield DurationInput(ps, lengths, ds)
                batch = []


def load_textgrid_wav(
    data_dir: Path, token_seq_len: int, batch_size, pad_wav_len, mode: str
):
    """load wav and textgrid files to memory."""
    tg_files = sorted(data_dir.glob("*.TextGrid"))
    random.Random(42).shuffle(tg_files)
    L = len(tg_files) * 95 // 100
    assert mode in ["train", "val", "gta"]
    phonemes = load_phonemes_set()
    if mode == "gta":
        tg_files = tg_files  # all files
    elif mode == "train":
        tg_files = tg_files[:L]
    elif mode == "val":
        tg_files = tg_files[L:]

    data = []
    for fn in tg_files:
        ps, ds = zip(*load_textgrid(fn))
        ps = [phonemes.index(p) for p in ps]
        l = len(ps)
        ps = pad_seq(ps, token_seq_len, 0)
        ds = pad_seq(ds, token_seq_len, 0)

        wav_file = data_dir / f"{fn.stem}.wav"
        sr, y = wavfile.read(wav_file)
        y = np.copy(y)
        start_time = 0
        for i, (phone_idx, duration) in enumerate(zip(ps, ds)):
            l = int(start_time * sr)
            end_time = start_time + duration
            r = int(end_time * sr)
            if i == len(ps) - 1:
                r = len(y)
            if phone_idx < len(FLAGS.special_phonemes):
                y[l:r] = 0
            start_time = end_time

        if len(y) > pad_wav_len:
            y = y[:pad_wav_len]

        # # normalize to match hifigan preprocessing
        # y = y.astype(np.float32)
        # y = y / np.max(np.abs(y))
        # y = y * 0.95
        # y = y * (2 ** 15)
        # y = y.astype(np.int16)

        wav_length = len(y)
        y = np.pad(y, (0, pad_wav_len - len(y)))
        data.append((fn.stem, ps, ds, l, y, wav_length))

    batch = []
    while True:
        random.shuffle(data)
        for idx, e in enumerate(data):
            batch.append(e)
            if len(batch) == batch_size or (mode == "gta" and idx == len(data) - 1):
                names, ps, ds, lengths, wavs, wav_lengths = zip(*batch)
                ps = np.array(ps, dtype=np.int32)
                ds = np.array(ds, dtype=np.float32)
                lengths = np.array(lengths, dtype=np.int32)
                wavs = np.array(wavs, dtype=np.int16)
                wav_lengths = np.array(wav_lengths, dtype=np.int32)
                if mode == "gta":
                    yield names, AcousticInput(ps, lengths, ds, wavs, wav_lengths, None)
                else:
                    yield AcousticInput(ps, lengths, ds, wavs, wav_lengths, None)
                batch = []
        if mode == "gta":
            assert len(batch) == 0
            break



import os
import sys
from underthesea.corpus.data import Sentence
from underthesea.model_fetcher import ModelFetcher
from underthesea.models.text_classifier import TextClassifier
from os.path import dirname

sys.path.insert(0, dirname(dirname(__file__)))
classifier = None


def sentiment(X):
    global classifier
    model_name = 'SA_GENERAL_V131'
    model_path = ModelFetcher.get_model_path(model_name)

    if not classifier:
        if not os.path.exists(model_path):
            ModelFetcher.download(model_name)
        classifier = TextClassifier.load(model_path)

    sentence = Sentence(X)
    classifier.predict(sentence)
    labels = sentence.labels
    try:
        label_map = {'POS': 'positive', 'NEG': 'negative'}
        label = label_map[labels[0]]
        return label
    except Exception:
        return None



"""
Underthesea Vietnamese IPA
"""
from collections import OrderedDict
from types import SimpleNamespace
import regex
from underthesea.pipeline.word_tokenize.regex_tokenize import VIETNAMESE_VOWELS_LOWER
from underthesea.utils.vietnamese_ipa_rules import codas, nuclei, onsets, onoffglides, onglides, offglides


class VIETNAMESE:
    """VIETNAMESE NAMESPACES"""

    LOWERS = SimpleNamespace(
        WITHOUT_DIACRITIC="aăâbcdđeêfghijklmnoôơpqrstuưvwxyz",
        HIGH_LEVEL="aăâeêioôơuưy",
        MID_FALLING="àằầèềìòồờùừỳ",
        RISING="áắấéếíóốớúứý",
        LOW_FALLING_RISING="ảẳẩẻểỉỏổởủửỷ",
        HIGH_FALLING_RISING_GLOTTALIZED="ãẵẫẽễĩõỗỡũữỹ",
        LOW_GLOTTALIZED="ạặậẹệịọộợụựỵ",
    )

    TONE = SimpleNamespace(
        HIGH_LEVEL="HIGH_LEVEL",  # e.g. ta, ba
        MID_FALLING="MID_FALLING",  # e.g. tà, bà
        RISING="RISING",  # e.g. tá, bá
        LOW_FALLING_RISING="LOW_FALLING_RISING",  # e.g. tả, bả
        HIGH_FALLING_RISING_GLOTTALIZED="HIGH_FALLING_RISING_GLOTTALIZED",  # e.g. tã, bã
        LOW_GLOTTALIZED="LOW_GLOTTALIZED",  # e.g. tạ, bạ
    )

    @classmethod
    def analyze_tone(cls, text: str):
        """

        Args:
            text (str): a syllable
        """
        hl = cls.LOWERS.HIGH_LEVEL
        mf = cls.LOWERS.MID_FALLING
        r = cls.LOWERS.RISING
        lfr = cls.LOWERS.LOW_FALLING_RISING
        hfrg = cls.LOWERS.HIGH_FALLING_RISING_GLOTTALIZED
        lg = cls.LOWERS.LOW_GLOTTALIZED
        d1 = mf + r + lfr + hfrg + lg
        d2 = hl + hl + hl + hl + hl
        tmp = text.maketrans(d1, d2)
        non_tone_letters = text.translate(tmp)

        tone = cls.TONE.HIGH_LEVEL
        for c in text:
            if c in mf:
                tone = cls.TONE.MID_FALLING
            elif c in r:
                tone = cls.TONE.RISING
            elif c in lfr:
                tone = cls.TONE.LOW_FALLING_RISING
            elif c in hfrg:
                tone = cls.TONE.HIGH_FALLING_RISING_GLOTTALIZED
            elif c in lg:
                tone = cls.TONE.LOW_GLOTTALIZED
        return non_tone_letters, tone


class Syllable:
    """
    Syllable class
    """
    def __init__(self, text):
        self.text = text
        non_tone_letters, tone = VIETNAMESE.analyze_tone(text)
        self.tone = tone
        a = "[aăâ]"
        o = "[oôơ]"
        u = "[uư]"
        double = f"oo|i{a}|iê|yê|y{a}|{u}{o}|{u}{a}|ay|ây|ua"
        v = r"(?P<V>[aăâeêuưyoôơi]|" + double + ")"
        vec = r"(?P<V>ê)"
        wu = r"(?P<w>[u])"
        vye = "(?P<V>yê)"
        vya = f"(?P<V>y{a})"
        consonants = "gi|qu|ch|gh|kh|ng|ngh|nh|ph|th|tr|[bcdđghklmnpqrstvx]"
        conda = consonants + "|[uio]"
        c1 = r"(?P<C1>" + consonants + ")?"
        c2 = r"(?P<C2>" + conda + ")?"
        patterns = [
            r"^" + c1 + "(?P<w>[u])(?P<V>[yâ])" + c2 + "$",
            r"^" + c1 + "(?P<w>[o])(?P<V>[eaă])" + c2 + "$",
            r"^" + c1 + v + c2 + "$",
            r"^" + c1 + wu + vec + c2 + "$",
            r"^" + c1 + wu + vye + c2 + "$",
            r"^" + c1 + wu + vya + c2 + "$",
        ]
        pattern = r"(" + "|".join(patterns) + ")"

        matched = regex.match(pattern, non_tone_letters)
        self.matched = matched
        if not matched:
            raise Exception(f"Text {text} not matched")

    def _util_reverse_dict(self, d):
        result = {}
        for k in d:
            for v in d[k]:
                result[v] = k
        return result

    # flake8: noqa: C901
    def generate_ipa(
            self, dialect: str = "north", eight: bool = False, tone: str = "number"
    ):
        """Generate ipa of the syllable

        Vietnamese syllabic structure (Trang 2022)
        syllable = onset + rhyme + tone
        rhyme = medial + nuclear vowel + (coda)

        Args:
            dialect (str): Either the `string` `"north"` or `"south"`. Default: `north`
            eight (boolean): If true, use eight tone format, else use six tone format. Default: `False`
            tone (str): Either the `string` `"ipa"` or `"number"`. Default: `number`

        Returns:
            A `string`. Represents ipa of the syllable
        """
        groups = self.matched.groupdict()

        map_w = {
            "o": "ʷ",
            "u": "ʷ"
        }

        c1, w, v, c2 = groups["C1"], groups["w"], groups["V"], groups["C2"]

        ipa_w = ""
        if w:
            ipa_w = map_w[w]

        nuclei.update(onglides)
        nuclei.update(offglides)
        nuclei.update(onoffglides)
        ipa_v = nuclei[v]
        if c1:
            ipa_c1 = onsets[c1]
        else:
            ipa_c1 = ""
            if v == "a":
                # ipa_v = "ɛ̆"
                ipa_v = "aː"

        if ipa_c1 == "":
            # ipa_c1 = "ʔ"
            ipa_c1 = ""

        if c2:
            ipa_c2 = codas[c2]
            if c2 in ["o", "i", "u", "y"]:
                if v == "a":
                    if c2 == "o":
                        # This rule apply in case ao -> ʔaw³³ not ɛ̆w³
                        # ipa_v = "a"
                        ipa_v = "aː"
                    elif c2 in ["u", "ă", "y"]:
                        ipa_v = "a"
                if v == "o" and c2 == "o":
                    ipa_v = "ↄ:"
                    ipa_c2 = ""
                if v == "y" and c2 == "u":
                    ipa_v = "i"
                if v == "u" and c2 == "y":
                    ipa_c2 = "i"
        else:
            ipa_c2 = ""

        map_tone_ipa = {
            VIETNAMESE.TONE.HIGH_LEVEL: "˧˧",
            VIETNAMESE.TONE.MID_FALLING: "˨˩",
            VIETNAMESE.TONE.LOW_FALLING_RISING: "˨˩",
            VIETNAMESE.TONE.HIGH_FALLING_RISING_GLOTTALIZED: "˧˥",
        }

        # vphon
        map_tone_number = {
            VIETNAMESE.TONE.HIGH_LEVEL: "³³",
            VIETNAMESE.TONE.MID_FALLING: "³²",
            VIETNAMESE.TONE.RISING: "²⁴",
            VIETNAMESE.TONE.LOW_FALLING_RISING: "³¹²",
            VIETNAMESE.TONE.HIGH_FALLING_RISING_GLOTTALIZED: "³ˀ⁵",
            VIETNAMESE.TONE.LOW_GLOTTALIZED: "²¹ˀ",
        }
        if tone == "number":
            map_t = map_tone_number
        else:
            map_t = map_tone_ipa
        ipa_t = map_t[self.tone]

        ons = ipa_c1
        cod = ipa_c2
        nuc = ipa_v
        ton = ipa_t

        # Deal with gi, giền and giêng
        if ons == "z" and nuc == "e":
            nuc = "iə"
        if ons == "ɣ" and nuc == "i":
            ons = "z"

        ##
        # Generate internal G2P representation
        ##
        if ons == "":
            ons = "ʔ"
        if ons == "ʂ":
            ons = "s"

        ##
        # Northern
        ##
        if dialect == "north":
            # Onset mergers
            if ons in ["j", "r"]:
                ons = "z"
            elif ons in ["c", "ʈ"]:
                ons = "tɕ"
            elif ons == "ʂ":
                ons = "s"

        if dialect == "south":
            if cod in ["ŋ", "k"] and nuc in ["u", "o", "ↄ"]:
                if cod == "ŋ":
                    cod = "ŋ͡m"
                elif cod == "k":
                    cod = "k͡p"

        # Capture vowel/coda interactions of ɛ/ɛː and e/eː
        if cod in ["ŋ", "k"]:
            if nuc == "ɛ":
                nuc = "ɛː"
            if nuc == "e":
                nuc = "eː"

        # Velar fronting
        if nuc == "aː":
            if cod == "c":
                nuc = "ɛ"
            if cod == "ɲ":
                nuc = "ɛ"

        # No surface palatal codas
        if cod in ["c", "ɲ"]:
            if cod == "c":
                cod = "k"
            if cod == "ɲ":
                cod = "ŋ"

        if not cod and nuc in ["aː", "əː"]:
            if nuc == "aː":
                nuc = "a"
            if nuc == "əː":
                nuc = "ə"

        if eight:
            if cod in ["p", "t", "k"]:
                if self.tone == VIETNAMESE.TONE.RISING:
                    ton = "⁴⁵"
                elif self.tone == VIETNAMESE.TONE.LOW_GLOTTALIZED:
                    ton = "²¹"
        ipa = ons + ipa_w + nuc + cod + ton
        return ipa


vietnamese_alphabet = (
        "aàáảãạ"
        + "ăằắẳẵặ"
        + "âầấẩẫậ"
        + "bcd"
        + "đ"
        + "eèéẻẽẹ"
        + "êềếểễệ"
        + "fgh"
        + "iìíỉĩị"
        + "jklmn"
        + "oòóỏõọ"
        + "ôồốổỗộ"
        + "ơờớởỡợ"
        + "pqrst"
        + "uùúủũụ"
        + "ưừứửữự"
        + "vwx"
        + "yỳýỷỹỵ"
        + "z"
)

vietnamese_alphabet_order = OrderedDict()
for i, character in enumerate(vietnamese_alphabet):
    vietnamese_alphabet_order[character] = i


def vietnamese_sort_key(s):
    return [vietnamese_alphabet_order[c] for c in s]



from os.path import join, dirname
from underthesea.util.singleton import Singleton
import pycrfsuite
import sys

if sys.version_info >= (3, 0):
    from .tagged_feature import word2features
else:
    from tagged_feature import word2features


@Singleton
class CRFChunkingPredictor:
    def __init__(self):
        self.model = pycrfsuite.Tagger()
        filepath = join(dirname(__file__), "chunk_crf_2017_10_12.bin")
        self.model.open(filepath)

        template = [
            "T[-2].lower", "T[-1].lower", "T[0].lower", "T[1].lower",
            "T[2].lower",
            "T[0].istitle", "T[-1].istitle", "T[1].istitle",
            # word unigram and bigram
            "T[-2]", "T[-1]", "T[0]", "T[1]", "T[2]",
            "T[-2,-1]", "T[-1,0]", "T[0,1]", "T[1,2]",
            # pos unigram and bigram
            "T[-2][1]", "T[-1][1]", "T[0][1]", "T[1][1]", "T[2][1]",
            "T[-2,-1][1]", "T[-1,0][1]", "T[0,1][1]", "T[1,2][1]",
            # chunk
            "T[-3][2]", "T[-2][2]", "T[-1][2]",
        ]
        self.transformer = TaggedTransformer(template)

    def predict(self, sentence, format=None):
        tokens = [(tokens[0], tokens[1], "X") for tokens in sentence]
        x = self.transformer.transform([tokens])[0][0]
        tags = self.model.tag(x)
        output = [(tokens[0], tokens[1], tag) for tokens, tag in
                  zip(sentence, tags)]
        return output


class TaggedTransformer:
    def __init__(self, template=None):
        self.template = template

    def transform(self, sentences):
        X = [self.sentence2features(s) for s in sentences]
        y = [self.sentence2labels(s) for s in sentences]
        return X, y

    def sentence2features(self, s):
        return [word2features(s, i, self.template) for i in range(len(s))]

    def sentence2labels(self, s):
        return [row[-1] for row in s]



"""
Underthesea
"""
# -*- coding: utf-8 -*-
import os
import sys
from functools import lru_cache


__author__ = """Vu Anh"""
__email__ = 'anhv.ict91@gmail.com'

# Check python version
try:
    version_info = sys.version_info
    if version_info < (3, 9, 0):
        raise RuntimeError("underthesea requires Python 3.9 or later")
except Exception:
    pass

###########################################################
# METADATA
###########################################################

# Version
try:
    version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
    with open(version_file, 'r') as infile:
        __version__ = infile.read().strip()
except NameError:
    __version__ = 'unknown (running code interactively?)'
except IOError as ex:
    __version__ = "unknown (%s)" % ex

###########################################################
# TOP-LEVEL MODULES
###########################################################
from .pipeline.sent_tokenize import sent_tokenize
from .pipeline.text_normalize import text_normalize
from .pipeline.word_tokenize import word_tokenize
from .pipeline.pos_tag import pos_tag
from .pipeline.chunking import chunk
from .pipeline.ner import ner

optional_imports = {
    'classify': 'underthesea.pipeline.classification',
    'sentiment': 'underthesea.pipeline.sentiment',
    'lang_detect': 'underthesea.pipeline.lang_detect',
    'dependency_parse': 'underthesea.pipeline.dependency_parse'
}


@lru_cache(maxsize=None)
def get_optional_import(module_name, object_name):
    try:
        module = __import__(module_name, fromlist=[object_name])
        return getattr(module, object_name)
    except ImportError:
        return None


for name, module in optional_imports.items():
    globals()[name] = get_optional_import(module, name)

__all__ = [
    'sent_tokenize',
    'text_normalize',
    'word_tokenize',
    'pos_tag',
    'chunk',
    'ner',
    'lang_detect',
    # 'classify',
    # 'sentiment',
    'dependency_parse'
]



import os
from datetime import timedelta, datetime
from pathlib import Path
from typing import Union
import torch.distributed as dist
import torch.nn as nn
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR
from underthesea.transforms.conll import CoNLL, progress_bar
from underthesea.models.dependency_parser import DependencyParser
from underthesea.utils import device, logger
from underthesea.utils.sp_common import pad, unk, bos
from underthesea.utils.sp_data import Dataset
from underthesea.utils.sp_embedding import Embedding
from underthesea.utils.sp_field import Field, SubwordField
from underthesea.utils.sp_metric import Metric, AttachmentMetric
from underthesea.utils.sp_parallel import DistributedDataParallel as DDP, is_master


class DependencyParserTrainer:
    def __init__(self, parser, corpus):
        self.parser = parser
        self.corpus = corpus

    # flake8: noqa: C901
    def train(
        self, base_path: Union[Path, str],
        fix_len=20,
        min_freq=2,
        buckets=1000,
        batch_size=5000,
        lr=2e-3,
        mu=.9,
        nu=.9,
        epsilon=1e-12,
        clip=5.0,
        decay=.75,
        decay_steps=5000,
        patience=100,
        max_epochs=10,
        wandb=None
    ):
        r"""
        Train any class that implement model interface

        Args:
            base_path (object): Main path to which all output during training is logged and models are saved
            max_epochs: Maximum number of epochs to train. Terminates training if this number is surpassed.
            patience:
            decay_steps:
            decay:
            clip:
            epsilon:
            nu:
            mu:
            lr:
            proj:
            tree:
            batch_size:
            buckets:
            min_freq:
            fix_len:


        """
        ################################################################################################################
        # BUILD
        ################################################################################################################
        feat = self.parser.feat
        embed = self.parser.embed
        os.makedirs(os.path.dirname(base_path), exist_ok=True)
        logger.info("Building the fields")
        WORD = Field('words', pad=pad, unk=unk, bos=bos, lower=True)
        if feat == 'char':
            FEAT = SubwordField('chars', pad=pad, unk=unk, bos=bos, fix_len=fix_len)
        elif feat == 'bert':
            from transformers import AutoTokenizer
            tokenizer = AutoTokenizer.from_pretrained(self.parser.bert)
            FEAT = SubwordField('bert',
                                pad=tokenizer.pad_token,
                                unk=tokenizer.unk_token,
                                bos=tokenizer.bos_token or tokenizer.cls_token,
                                fix_len=fix_len,
                                tokenize=tokenizer.tokenize)
            FEAT.vocab = tokenizer.get_vocab()
        else:
            FEAT = Field('tags', bos=bos)

        ARC = Field('arcs', bos=bos, use_vocab=False, fn=CoNLL.get_arcs)
        REL = Field('rels', bos=bos)
        if feat in ('char', 'bert'):
            transform = CoNLL(FORM=(WORD, FEAT), HEAD=ARC, DEPREL=REL)
        else:
            transform = CoNLL(FORM=WORD, CPOS=FEAT, HEAD=ARC, DEPREL=REL)

        train = Dataset(transform, self.corpus.train)
        WORD.build(train, min_freq, (Embedding.load(embed, unk) if self.parser.embed else None))
        FEAT.build(train)
        REL.build(train)
        n_words = WORD.vocab.n_init
        n_feats = len(FEAT.vocab)
        n_rels = len(REL.vocab)
        pad_index = WORD.pad_index
        unk_index = WORD.unk_index
        feat_pad_index = FEAT.pad_index
        parser = DependencyParser(
            n_words=n_words,
            n_feats=n_feats,
            n_rels=n_rels,
            pad_index=pad_index,
            unk_index=unk_index,
            feat_pad_index=feat_pad_index,
            transform=transform,
            feat=self.parser.feat,
            bert=self.parser.bert
        )
        # word_field_embeddings = self.parser.embeddings[0]
        # word_field_embeddings.n_vocab = 100
        parser.embeddings = self.parser.embeddings
        # parser.embeddings[0] = word_field_embeddings
        parser.load_pretrained(WORD.embed).to(device)

        ################################################################################################################
        # TRAIN
        ################################################################################################################
        if wandb:
            wandb.watch(parser)
        parser.transform.train()
        if dist.is_initialized():
            batch_size = batch_size // dist.get_world_size()
        logger.info('Loading the data')
        train = Dataset(parser.transform, self.corpus.train)
        dev = Dataset(parser.transform, self.corpus.dev)
        test = Dataset(parser.transform, self.corpus.test)
        train.build(batch_size, buckets, True, dist.is_initialized())
        dev.build(batch_size, buckets)
        test.build(batch_size, buckets)
        logger.info(f"\n{'train:':6} {train}\n{'dev:':6} {dev}\n{'test:':6} {test}\n")
        logger.info(f'{parser}')
        if dist.is_initialized():
            parser = DDP(parser, device_ids=[dist.get_rank()], find_unused_parameters=True)

        optimizer = Adam(parser.parameters(), lr, (mu, nu), epsilon)
        scheduler = ExponentialLR(optimizer, decay ** (1 / decay_steps))

        elapsed = timedelta()
        best_e, best_metric = 1, Metric()

        for epoch in range(1, max_epochs + 1):
            start = datetime.now()
            logger.info(f'Epoch {epoch} / {max_epochs}:')

            parser.train()

            bar = progress_bar(train.loader)
            metric = AttachmentMetric()
            for words, feats, arcs, rels in bar:
                optimizer.zero_grad()

                mask = words.ne(parser.WORD.pad_index)
                # ignore the first token of each sentence
                mask[:, 0] = 0
                s_arc, s_rel = parser.forward(words, feats)
                loss = parser.forward_loss(s_arc, s_rel, arcs, rels, mask)
                loss.backward()
                nn.utils.clip_grad_norm_(parser.parameters(), clip)
                optimizer.step()
                scheduler.step()

                arc_preds, rel_preds = parser.decode(s_arc, s_rel, mask)
                # ignore all punctuation if not specified
                if not self.parser.args['punct']:
                    mask &= words.unsqueeze(-1).ne(parser.puncts).all(-1)
                metric(arc_preds, rel_preds, arcs, rels, mask)
                bar.set_postfix_str(f'lr: {scheduler.get_last_lr()[0]:.4e} - loss: {loss:.4f} - {metric}')

            dev_loss, dev_metric = parser.evaluate(dev.loader)
            logger.info(f"{'dev:':6} - loss: {dev_loss:.4f} - {dev_metric}")
            test_loss, test_metric = parser.evaluate(test.loader)
            logger.info(f"{'test:':6} - loss: {test_loss:.4f} - {test_metric}")
            if wandb:
                wandb.log({"test_loss": test_loss})
                wandb.log({"test_metric_uas": test_metric.uas})
                wandb.log({"test_metric_las": test_metric.las})

            t = datetime.now() - start
            # save the model if it is the best so far
            if dev_metric > best_metric:
                best_e, best_metric = epoch, dev_metric
                if is_master():
                    parser.save(base_path)
                logger.info(f'{t}s elapsed (saved)\n')
            else:
                logger.info(f'{t}s elapsed\n')
            elapsed += t
            if epoch - best_e >= patience:
                break
        loss, metric = parser.load(base_path).evaluate(test.loader)

        logger.info(f'Epoch {best_e} saved')
        logger.info(f"{'dev:':6} - {best_metric}")
        logger.info(f"{'test:':6} - {metric}")
        logger.info(f'{elapsed}s elapsed, {elapsed / epoch}s/epoch')



from os.path import dirname, join

from underthesea.utils.col_dictionary import Dictionary

PROJECT_FOLDER = dirname(dirname(dirname(__file__)))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")
DICTIONARY_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL", "dictionary")
DICTIONARY_FILE = join(DICTIONARY_FOLDER, "202108.yaml")

if __name__ == '__main__':
    dictionary = Dictionary.load(DICTIONARY_FILE)
    dictionary.describe()
    for pos in ["verb", "noun", "adjective", "pronoun", "preposition"]:
        dictionary.import_words(join(DICTIONARY_FOLDER, "data", f"words_{pos}_candidates.xlsx"))
    dictionary.describe()
    dictionary.save(join(DICTIONARY_FOLDER, "202108.yaml"))



# -*- coding: utf-8 -*-


def sentiment(X, domain='general'):
    """
    Sentiment Analysis

    Parameters
    ==========

    X: str
        raw sentence
    domain: str
        domain of text (bank or general). Default: `general`

    Returns
    =======
        Text: Text of input sentence
        Labels: Sentiment of sentence

    Examples
    --------

        >>> from underthesea import sentiment
        >>> sentence = "Chuyen tiền k nhận Dc tiên"
        >>> sentiment(sentence, domain='bank')
        [MONEY_TRANSFER#negative (1.0)]
    """
    if X == "":
        return None
    if domain == 'general':
        from underthesea.pipeline.sentiment.general import sentiment
        return sentiment(X)
    if domain == 'bank':
        from underthesea.pipeline.sentiment.bank import sentiment
        return sentiment(X)



# ===========================
# token syntax
# ===========================
#         _ row 1
#        /  _ row 2
#       /  /  _ column
#      /  /  /
#    T[0,2][0]
#          .is_digit
#            \_ function
#
# ===========================
# sample tagged sentence
# ===========================
# this     A
# is       B
# a        C
# sample   D
# sentence E
#

import re

from os.path import join, dirname

from underthesea.corpus import DictionaryLoader

words = DictionaryLoader(join(dirname(__file__), "Viet74K.txt")).words
lower_words = set([word.lower() for word in words])


def text_lower(word):
    return word.lower()


def text_isdigit(word):
    return str(word.isdigit())


def text_isallcap(word):
    for letter in word:
        if not letter.istitle():
            return False
    return True


def text_istitle(word):
    if len(word) == 0:
        return False
    try:
        titles = [s[0] for s in word.split(" ")]
        for token in titles:
            if token[0].istitle() is False:
                return False
        return True
    except Exception:
        return False


def text_is_in_dict(word):
    return str(word.lower() in lower_words)


def apply_function(name, word):
    functions = {
        "lower": text_lower,
        "istitle": text_istitle,
        "isallcap": text_isallcap,
        "isdigit": text_isdigit,
        "is_in_dict": text_is_in_dict
    }
    return functions[name](word)


def template2features(sent, i, token_syntax, debug=True):
    """
    :type token: object
    """
    columns = []
    for j in range(len(sent[0])):
        columns.append([t[j] for t in sent])
    matched = re.match(
        "T\[(?P<index1>\-?\d+)(\,(?P<index2>\-?\d+))?\](\[(?P<column>.*)\])?(\.(?P<function>.*))?",
        token_syntax)
    column = matched.group("column")
    column = int(column) if column else 0
    index1 = int(matched.group("index1"))
    index2 = matched.group("index2")
    index2 = int(index2) if index2 else None
    func = matched.group("function")
    if debug:
        prefix = "%s=" % token_syntax
    else:
        prefix = ""
    if i + index1 < 0:
        return ["%sBOS" % prefix]
    if i + index1 >= len(sent):
        return ["%sEOS" % prefix]
    if index2 is not None:
        if i + index2 >= len(sent):
            return ["%sEOS" % prefix]
        word = " ".join(columns[column][i + index1: i + index2 + 1])
    else:
        word = sent[i + index1][column]
    if func is not None:
        result = apply_function(func, word)
    else:
        result = word
    return ["%s%s" % (prefix, result)]


def word2features(sent, i, template):
    features = []
    for token in template:
        features.extend(template2features(sent, i, token))
    return features


def sent2features(sentence, template):
    """ extract features in a sentence

    :type sentence: list of token, each token is a list of tag
    """
    return [word2features(sentence, i, template) for i in range(len(sentence))]


def sent2labels(sentence):
    """ extract labels in a sentence

    :type sentence: list of token, each token is a list of tag
    :return a list contains labels in sentence

    Example:
        sentence = [("Have", "V"), ("a", "D"), ("nice", A"), (day, "N")
    """
    return [label for token, label in sentence]



from os import makedirs
from os.path import basename, join, exists
from shutil import copyfile, rmtree

from underthesea.corpus.validate_corpus import validate_corpus_exist
from underthesea.feature_engineering.text import Text
from underthesea.file_utils import DATASETS_FOLDER


def revise_vlsp2013_wtk_dataset(source_file, dest_file):
    copyfile(source_file, dest_file)
    with open(source_file) as f:
        content = Text(f.read())
        sentences = content.split("\n\n")
    if basename(source_file) == "train.txt":
        corpus_id = "train"
    else:
        corpus_id = "test"
    last_index = len(sentences) - 1
    with open(dest_file, "w") as f:
        for i, sentence in enumerate(sentences):
            nodes = sentence.split("\n")
            nodes = [node.split("\t") for node in nodes]
            send_id = f"# sent_id = {corpus_id}-s{i + 1}\n"
            text = " ".join([node[0] for node in nodes])
            text = f"# text = {text}\n"
            content = send_id + text + sentence
            if i != last_index:
                content += "\n\n"
            else:
                content += "\n"
            f.write(content)


def revise_corpus_vlsp2013_wtk():
    source_corpus = "VLSP2013-WTK"
    validate_corpus_exist(source_corpus)
    revise_version = 2
    dest_corpus = f"{source_corpus}-R{revise_version}"
    source_folder = join(DATASETS_FOLDER, source_corpus)
    dest_folder = join(DATASETS_FOLDER, dest_corpus)
    if exists(dest_folder):
        rmtree(dest_folder)
    makedirs(dest_folder)
    files = ["train.txt", "test.txt"]
    for file in files:
        revise_vlsp2013_wtk_dataset(join(source_folder, file), join(dest_folder, file))



import os
import shutil
from os.path import dirname, join

from underthesea.corpus.util import FolderStructure
from underthesea.file_utils import UNDERTHESEA_FOLDER
from pathlib import Path


def process_train_data(train_data_file, output_file, label):
    with open(train_data_file) as f:
        text = f.read()
        text = text.strip()
        sentences = text.split("\n\n")
    with open(output_file, "a") as f:
        for s in sentences:
            f.write(f"__label__{label} {s}\n")


class VLSP2016SACorpus:
    data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "vlsp2016_sa"
    required_folder = True
    required_file = False

    @staticmethod
    def reset_folder():
        try:
            shutil.rmtree(VLSP2016SACorpus.data_folder)
        except Exception:
            pass
        os.makedirs(VLSP2016SACorpus.data_folder)

    @staticmethod
    def import_data(input_data_path: str):
        if VLSP2016SACorpus.required_folder and not Path(input_data_path).is_dir():
            print(f"Input path must be a folder but your input is {input_data_path}")
            raise SystemExit("")
        if VLSP2016SACorpus.required_file and not Path(input_data_path).is_file():
            print(f"Input path must be a file but your input is {input_data_path}")
            raise SystemExit("")
        raw_sample = join(dirname(dirname(__file__)), "../../corpus/data", "vlsp2016_sa_raw_sample")
        FolderStructure.check_structure(raw_sample, input_data_path)
        VLSP2016SACorpus.reset_folder()
        input_data_folder = Path(input_data_path)
        train_input_data_folder = input_data_folder / "SA2016-training_data"
        output_train_file = VLSP2016SACorpus.data_folder / "train.txt"
        with open(output_train_file, "w") as f:
            f.write("")
        process_train_data(train_input_data_folder / "SA-training_positive.txt", output_train_file, "POS")
        process_train_data(train_input_data_folder / "SA-training_neutral.txt", output_train_file, "NEU")
        process_train_data(train_input_data_folder / "SA-training_negative.txt", output_train_file, "NEG")

        # Preprocess Test Data
        test_sentences = []
        with open(input_data_folder / "SA2016-TestData-Ans" / "test_raw_ANS.txt", "r") as f:

            for i, line in enumerate(f):
                if i % 2 == 0:
                    text = line.strip()
                else:
                    label = line.strip()
                    sentence = f"__label__{label} {text}"
                    test_sentences.append(sentence)
        output_test_file = VLSP2016SACorpus.data_folder / "test.txt"
        with open(output_test_file, "w") as f:
            content = "\n".join(test_sentences)
            f.write(content + "\n")

        print(f"[Underthesea] Corpus `VLSP2016_SA` is imported in {VLSP2016SACorpus.data_folder}")



import sys

import unicodedata


def Text(text):
    """ provide a wrapper for python string
    map byte to str (python 3)
    map str to unicode (python 2)
    all string in utf-8 encoding
    normalize string to NFC
    """
    if not is_unicode(text):
        text = text.decode("utf-8")
    text = unicodedata.normalize("NFC", text)
    return text


def is_unicode(text):
    if sys.version_info >= (3, 0):
        unicode_type = str
    return type(text) is unicode_type



import os
import re
import shutil
from enum import Enum
from os.path import dirname, join
from typing import Union, List
import yaml
from underthesea.corpus import Corpus
from underthesea.corpus.categorized_corpus import CategorizedCorpus
from underthesea.corpus.data import Sentence, Label
from underthesea.file_utils import cached_path, UNDERTHESEA_FOLDER
from pathlib import Path
import zipfile

from underthesea.utils import print_table

MISS_URL_ERROR = (
    "Caution:\n  With closed license dataset, you must provide URL to download"
)
SAMPLE_CACHE_ROOT = Path(__file__).parent.absolute() / "data"

CD = dirname(__file__)
with open(join(CD, "datasets.yaml")) as f:
    REPO = yaml.safe_load(f)


class NLPData(Enum):
    AIVIVN2019_SA = "aivivn2019_sa"
    AIVIVN2019_SA_SAMPLE = "aivivn2019_sa_sample"
    UTS2017_BANK_SA = "uts2017_bank_sa"
    UTS2017_BANK_TC = "uts2017_bank_tc"
    UTS2017_BANK_SA_SAMPLE = "uts2017_bank_sa_sample"
    VLSP2016_SA = "vlsp2016_sa"
    VNTC = "VNTC"


class DataFetcher:
    @staticmethod
    def download_raw_file_to_cache(repo_data):
        url = repo_data["url"]
        url_filename = repo_data["url_filename"]
        cache_dir = repo_data["cache_dir"]
        filepath = repo_data["filepath"]
        cached_path(url, cache_dir=cache_dir)
        shutil.move(
            Path(UNDERTHESEA_FOLDER) / cache_dir / url_filename,
            Path(UNDERTHESEA_FOLDER) / cache_dir / filepath,
        )

    @staticmethod
    def download_zip_file_to_cache(repo_data):
        url = repo_data["url"]
        cache_dir = repo_data["cache_dir"]
        url_filename = repo_data["url_filename"]
        cached_path(url, cache_dir=cache_dir)
        filepath = Path(UNDERTHESEA_FOLDER) / cache_dir / url_filename
        cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
        with zipfile.ZipFile(filepath) as zip:
            zip.extractall(cache_folder)
        os.remove(filepath)

    @staticmethod
    def download_data(data, url):
        if data not in REPO:
            print(f"No matching distribution found for '{data}'")
            return
        repo_data = REPO[data]
        if "url" in repo_data:
            url = repo_data["url"]
        filepath = REPO[data]["filepath"]
        cache_dir = REPO[data]["cache_dir"]
        filepath = Path(UNDERTHESEA_FOLDER) / cache_dir / filepath
        if Path(filepath).exists():
            print(f"Resource {data} is already existed in: {filepath}")
            return

        if data in set(["VNESES", "VNTQ_SMALL", "VNTQ_BIG"]):
            DataFetcher.download_raw_file_to_cache(repo_data)

        zip_datasets = [
            "VNTC",
            "VLSP2013-WTK",
            "VLSP2013-POS",
            "VTB-CHUNK",
            "VLSP2016-NER",
            "VLSP2018-NER",
            "AIVIVN2019-SA",
            "VLSP2016-SA",
            "VLSP2018-SA",
            "UTS2017-BANK",
            "DI_Vietnamese-UVD",
            "CP_Vietnamese-UNC",
            "SE_Vietnamese-UBS",
            "UIT_ABSA_RESTAURANT",
            "UIT_ABSA_HOTEL",
            "CP_Vietnamese_VLC_v2_2022",
        ]
        if data in set(zip_datasets):
            if repo_data["license"] == "Close":
                if not url:
                    print(f"\n{MISS_URL_ERROR}")
                    return
                repo_data["url"] = url
            DataFetcher.download_zip_file_to_cache(repo_data)
        print(f"Resource {data} is downloaded in {filepath} folder")

    @staticmethod
    def list(all):
        datasets = []
        for key in REPO:
            name = key
            type = REPO[key]["type"]
            license = REPO[key]["license"]
            year = REPO[key]["year"]
            directory = REPO[key]["cache_dir"]
            if not all:
                if license == "Close":
                    continue
            if license == "Close":
                license = "Close*"
            datasets.append([name, type, license, year, directory])
        datasets = list(
            sorted(datasets, key=lambda x: (x[3], x[1], x[0]), reverse=True)
        )
        print_table(datasets, headers=["Name", "Type", "License", "Year", "Directory"])

        if all:
            print(f"\n{MISS_URL_ERROR}")

    @staticmethod
    def remove(data):
        if data not in REPO:
            print(f"No matching distribution found for '{data}'")
            return
        dataset = REPO[data]
        cache_dir = Path(UNDERTHESEA_FOLDER) / dataset["cache_dir"]
        if cache_dir.is_dir():
            shutil.rmtree(cache_dir)
        print(f"Dataset {data} is removed.")

    @staticmethod
    def load_corpus(corpus_id: Union[NLPData, str]) -> Corpus:
        if corpus_id == NLPData.AIVIVN2019_SA:
            data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "aivivn2019_sa"
            return DataFetcher.load_classification_corpus(data_folder)

        if corpus_id == NLPData.AIVIVN2019_SA_SAMPLE:
            data_folder = SAMPLE_CACHE_ROOT / "aivivn2019_sa_sample"
            return DataFetcher.load_classification_corpus(data_folder)

        if corpus_id == NLPData.UTS2017_BANK_SA:
            data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "uts2017_bank"
            return DataFetcher.load_classification_corpus(data_folder)

        if corpus_id == NLPData.UTS2017_BANK_SA_SAMPLE:
            data_folder = SAMPLE_CACHE_ROOT / "uts2017_bank_sample"
            return DataFetcher.load_classification_corpus(data_folder)

        if corpus_id == NLPData.UTS2017_BANK_TC:
            data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "uts2017_bank"
            corpus = DataFetcher.load_classification_corpus(data_folder)
            return DataFetcher.__exact_aspect_labels(corpus)

        if corpus_id == NLPData.VLSP2016_SA:
            data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "vlsp2016_sa"
            corpus = DataFetcher.load_classification_corpus(data_folder)
            return DataFetcher.__exact_aspect_labels(corpus)

        if corpus_id == NLPData.VNTC:
            data_folder = Path(UNDERTHESEA_FOLDER) / "datasets" / "VNTC"
            corpus = DataFetcher.load_classification_corpus(data_folder)
            return DataFetcher.__exact_aspect_labels(corpus)

    @staticmethod
    def __exact_aspect_labels(corpus: CategorizedCorpus):
        def extract(data: List[Sentence]):
            for sentence in data:
                labels = []
                for label in sentence.labels:
                    value = label.value.split("#")[0]
                    label.value = value
                    labels.append(label)
                sentence.labels = labels
            return data

        train = extract(corpus.train)
        dev = extract(corpus.dev)
        test = extract(corpus.test)
        output_corpus = CategorizedCorpus(train, dev, test)
        return output_corpus

    @staticmethod
    def load_classification_corpus(data_folder) -> CategorizedCorpus:
        train_file = data_folder / "train.txt"
        dev_file = data_folder / "dev.txt"
        test_file = data_folder / "test.txt"
        sentences_train: List[Sentence] = DataFetcher.read_text_classification_file(
            train_file
        )
        if dev_file.is_file():
            sentences_dev: List[Sentence] = DataFetcher.read_text_classification_file(
                dev_file
            )
        else:
            sentences_train, sentences_dev = DataFetcher.__sample(sentences_train)
        sentences_test: List[Sentence] = DataFetcher.read_text_classification_file(
            test_file
        )
        corpus = CategorizedCorpus(sentences_train, sentences_dev, sentences_test)
        return corpus

    @staticmethod
    def __sample(data: List[Sentence], percentage: float = 0.1):
        import random

        random.shuffle(data)
        index = int(len(data) * percentage)
        a = data[:-index]
        b = data[-index:]
        return a, b

    @staticmethod
    def read_text_classification_file(path_to_file) -> List[Sentence]:
        sentences = []
        with open(path_to_file) as f:
            lines = f.read().splitlines()
            for line in lines:
                label_pattern = r"__label__(?P<label>[\w#]+)"
                labels = re.findall(label_pattern, line)
                labels = [Label(label) for label in labels]
                text = re.sub(label_pattern, "", line)
                s = Sentence(text, labels)
                sentences.append(s)
        return sentences

    @staticmethod
    # WIP
    def import_data(corpus_id: str, input_data_path: str):
        if corpus_id not in REPO:
            print(f"No matching distribution found for '{corpus_id}'")
            return
        if corpus_id == "VLSP2016_SA":
            from underthesea.datasets.vlsp2016_sa.vlsp2016_sa_corpus import (
                VLSP2016SACorpus,
            )

            VLSP2016SACorpus.import_data(input_data_path)



# flake8: noqa

import math
import os
import random
import torch
import torch.utils.data
import numpy as np
from librosa.util import normalize
from scipy.io.wavfile import read
from librosa.filters import mel as librosa_mel_fn

MAX_WAV_VALUE = 32768.0


def load_wav(full_path):
    sampling_rate, data = read(full_path)
    return data, sampling_rate


def dynamic_range_compression(x, C=1, clip_val=1e-5):
    return np.log(np.clip(x, a_min=clip_val, a_max=None) * C)


def dynamic_range_decompression(x, C=1):
    return np.exp(x) / C


def dynamic_range_compression_torch(x, C=1, clip_val=1e-5):
    return torch.log(torch.clamp(x, min=clip_val) * C)


def dynamic_range_decompression_torch(x, C=1):
    return torch.exp(x) / C


def spectral_normalize_torch(magnitudes):
    output = dynamic_range_compression_torch(magnitudes)
    return output


def spectral_de_normalize_torch(magnitudes):
    output = dynamic_range_decompression_torch(magnitudes)
    return output


mel_basis = {}
hann_window = {}


def mel_spectrogram(
    y, n_fft, num_mels, sampling_rate, hop_size, win_size, fmin, fmax, center=False
):
    if torch.min(y) < -1.0:
        print("min value is ", torch.min(y))
    if torch.max(y) > 1.0:
        print("max value is ", torch.max(y))

    global mel_basis, hann_window
    if fmax not in mel_basis:
        mel = librosa_mel_fn(sampling_rate, n_fft, num_mels, fmin, fmax)
        mel_basis[str(fmax) + "_" + str(y.device)] = (
            torch.from_numpy(mel).float().to(y.device)
        )
        hann_window[str(y.device)] = torch.hann_window(win_size).to(y.device)

    y = torch.nn.functional.pad(
        y.unsqueeze(1),
        (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)),
        mode="reflect",
    )
    y = y.squeeze(1)

    spec = torch.stft(
        y,
        n_fft,
        hop_length=hop_size,
        win_length=win_size,
        window=hann_window[str(y.device)],
        center=center,
        pad_mode="reflect",
        normalized=False,
        onesided=True,
    )

    spec = torch.sqrt(spec.pow(2).sum(-1) + (1e-9))

    spec = torch.matmul(mel_basis[str(fmax) + "_" + str(y.device)], spec)
    spec = spectral_normalize_torch(spec)

    return spec


def get_dataset_filelist(a):
    with open(a.input_training_file, "r", encoding="utf-8") as fi:
        training_files = [
            os.path.join(a.input_wavs_dir, x.split("|")[0] + ".wav")
            for x in fi.read().split("\n")
            if len(x) > 0
        ]

    with open(a.input_validation_file, "r", encoding="utf-8") as fi:
        validation_files = [
            os.path.join(a.input_wavs_dir, x.split("|")[0] + ".wav")
            for x in fi.read().split("\n")
            if len(x) > 0
        ]
    return training_files, validation_files


class MelDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        training_files,
        segment_size,
        n_fft,
        num_mels,
        hop_size,
        win_size,
        sampling_rate,
        fmin,
        fmax,
        split=True,
        shuffle=True,
        n_cache_reuse=1,
        device=None,
        fmax_loss=None,
        fine_tuning=False,
        base_mels_path=None,
    ):
        self.audio_files = training_files
        random.seed(1234)
        if shuffle:
            random.shuffle(self.audio_files)
        self.segment_size = segment_size
        self.sampling_rate = sampling_rate
        self.split = split
        self.n_fft = n_fft
        self.num_mels = num_mels
        self.hop_size = hop_size
        self.win_size = win_size
        self.fmin = fmin
        self.fmax = fmax
        self.fmax_loss = fmax_loss
        self.cached_wav = None
        self.n_cache_reuse = n_cache_reuse
        self._cache_ref_count = 0
        self.device = device
        self.fine_tuning = fine_tuning
        self.base_mels_path = base_mels_path

    def __getitem__(self, index):
        filename = self.audio_files[index]
        if self._cache_ref_count == 0:
            audio, sampling_rate = load_wav(filename)
            audio = audio / MAX_WAV_VALUE
            if not self.fine_tuning:
                audio = normalize(audio) * 0.95
            self.cached_wav = audio
            if sampling_rate != self.sampling_rate:
                raise ValueError(
                    "{} SR doesn't match target {} SR".format(
                        sampling_rate, self.sampling_rate
                    )
                )
            self._cache_ref_count = self.n_cache_reuse
        else:
            audio = self.cached_wav
            self._cache_ref_count -= 1

        audio = torch.FloatTensor(audio)
        audio = audio.unsqueeze(0)

        if not self.fine_tuning:
            if self.split:
                if audio.size(1) >= self.segment_size:
                    max_audio_start = audio.size(1) - self.segment_size
                    audio_start = random.randint(0, max_audio_start)
                    audio = audio[:, audio_start: audio_start + self.segment_size]
                else:
                    audio = torch.nn.functional.pad(
                        audio, (0, self.segment_size - audio.size(1)), "constant"
                    )

            mel = mel_spectrogram(
                audio,
                self.n_fft,
                self.num_mels,
                self.sampling_rate,
                self.hop_size,
                self.win_size,
                self.fmin,
                self.fmax,
                center=False,
            )
        else:
            mel = np.load(
                os.path.join(
                    self.base_mels_path,
                    os.path.splitext(os.path.split(filename)[-1])[0] + ".npy",
                )
            )
            mel = torch.from_numpy(mel)

            if len(mel.shape) < 3:
                mel = mel.unsqueeze(0)

            if self.split:
                frames_per_seg = math.ceil(self.segment_size / self.hop_size)

                if audio.size(1) >= self.segment_size:
                    mel_start = random.randint(0, mel.size(2) - frames_per_seg - 1)
                    mel = mel[:, :, mel_start: mel_start + frames_per_seg]
                    audio = audio[
                        :,
                        mel_start *
                        self.hop_size: (mel_start + frames_per_seg) *
                        self.hop_size,
                    ]
                else:
                    mel = torch.nn.functional.pad(
                        mel, (0, frames_per_seg - mel.size(2)), "constant"
                    )
                    audio = torch.nn.functional.pad(
                        audio, (0, self.segment_size - audio.size(1)), "constant"
                    )

        mel_loss = mel_spectrogram(
            audio,
            self.n_fft,
            self.num_mels,
            self.sampling_rate,
            self.hop_size,
            self.win_size,
            self.fmin,
            self.fmax_loss,
            center=False,
        )

        return (mel.squeeze(), audio.squeeze(0), filename, mel_loss.squeeze())

    def __len__(self):
        return len(self.audio_files)



# -*- coding: utf-8 -*-

from collections import Counter


class Metric(object):

    def __lt__(self, other):
        return self.score < other

    def __le__(self, other):
        return self.score <= other

    def __ge__(self, other):
        return self.score >= other

    def __gt__(self, other):
        return self.score > other

    @property
    def score(self):
        return 0.


class AttachmentMetric(Metric):

    def __init__(self, eps=1e-8):
        super().__init__()

        self.eps = eps

        self.n = 0.0
        self.n_ucm = 0.0
        self.n_lcm = 0.0
        self.total = 0.0
        self.correct_arcs = 0.0
        self.correct_rels = 0.0

    def __repr__(self):
        s = f"UCM: {self.ucm:6.2%} LCM: {self.lcm:6.2%} "
        s += f"UAS: {self.uas:6.2%} LAS: {self.las:6.2%}"
        return s

    def __call__(self, arc_preds, rel_preds, arc_golds, rel_golds, mask):
        lens = mask.sum(1)
        arc_mask = arc_preds.eq(arc_golds) & mask
        rel_mask = rel_preds.eq(rel_golds) & arc_mask
        arc_mask_seq, rel_mask_seq = arc_mask[mask], rel_mask[mask]

        self.n += len(mask)
        self.n_ucm += arc_mask.sum(1).eq(lens).sum().item()
        self.n_lcm += rel_mask.sum(1).eq(lens).sum().item()

        self.total += len(arc_mask_seq)
        self.correct_arcs += arc_mask_seq.sum().item()
        self.correct_rels += rel_mask_seq.sum().item()

    @property
    def score(self):
        return self.las

    @property
    def ucm(self):
        return self.n_ucm / (self.n + self.eps)

    @property
    def lcm(self):
        return self.n_lcm / (self.n + self.eps)

    @property
    def uas(self):
        return self.correct_arcs / (self.total + self.eps)

    @property
    def las(self):
        return self.correct_rels / (self.total + self.eps)


class BracketMetric(Metric):

    def __init__(self, eps=1e-8):
        super().__init__()

        self.n = 0.0
        self.n_ucm = 0.0
        self.n_lcm = 0.0
        self.utp = 0.0
        self.ltp = 0.0
        self.pred = 0.0
        self.gold = 0.0
        self.eps = eps

    def __call__(self, preds, golds):
        for pred, gold in zip(preds, golds):
            upred = Counter([(i, j) for i, j, label in pred])
            ugold = Counter([(i, j) for i, j, label in gold])
            utp = list((upred & ugold).elements())
            lpred = Counter(pred)
            lgold = Counter(gold)
            ltp = list((lpred & lgold).elements())
            self.n += 1
            self.n_ucm += len(utp) == len(pred) == len(gold)
            self.n_lcm += len(ltp) == len(pred) == len(gold)
            self.utp += len(utp)
            self.ltp += len(ltp)
            self.pred += len(pred)
            self.gold += len(gold)

    def __repr__(self):
        s = f"UCM: {self.ucm:6.2%} LCM: {self.lcm:6.2%} "
        s += f"UP: {self.up:6.2%} UR: {self.ur:6.2%} UF: {self.uf:6.2%} "
        s += f"LP: {self.lp:6.2%} LR: {self.lr:6.2%} LF: {self.lf:6.2%}"

        return s

    @property
    def score(self):
        return self.lf

    @property
    def ucm(self):
        return self.n_ucm / (self.n + self.eps)

    @property
    def lcm(self):
        return self.n_lcm / (self.n + self.eps)

    @property
    def up(self):
        return self.utp / (self.pred + self.eps)

    @property
    def ur(self):
        return self.utp / (self.gold + self.eps)

    @property
    def uf(self):
        return 2 * self.utp / (self.pred + self.gold + self.eps)

    @property
    def lp(self):
        return self.ltp / (self.pred + self.eps)

    @property
    def lr(self):
        return self.ltp / (self.gold + self.eps)

    @property
    def lf(self):
        return 2 * self.ltp / (self.pred + self.gold + self.eps)


class SpanMetric(Metric):

    def __init__(self, eps=1e-5):
        super(SpanMetric, self).__init__()

        self.tp = 0.0
        self.pred = 0.0
        self.gold = 0.0
        self.eps = eps

    def __call__(self, preds, golds):
        for pred, gold in zip(preds, golds):
            pred, gold = set(pred), set(gold)
            self.tp += len(pred & gold)
            self.pred += len(pred)
            self.gold += len(gold)

    def __repr__(self):
        return f"P: {self.p:6.2%} R: {self.r:6.2%} F: {self.f:6.2%}"

    @property
    def score(self):
        return self.f

    @property
    def p(self):
        return self.tp / (self.pred + self.eps)

    @property
    def r(self):
        return self.tp / (self.gold + self.eps)

    @property
    def f(self):
        return 2 * self.p * self.r / (self.p + self.r + self.eps)



from functools import reduce

from underthesea.transformer import Transformer


class WordVectorTransformer(Transformer):
    """ Convert a collection of raw documents to a matrix of text-index
    features

    keep the order of word in document
    index start with 1, index 0 is default padding_value

    Parameters
    ----------

    embedding_size: size of embedding word to vector
    """

    def __init__(self, embedding_size=100, padding=False, padding_value=0,
                 unknown_value=1):
        self.vocab = []
        self.embedding_size = embedding_size
        self.padding = padding
        self.padding_value = padding_value
        self.unknown_value = unknown_value

        self.max_length = 0
        self.ns = 2  # number of special values

    def get_vocab_size(self):
        # Add 2 tokens for
        # - padding_value (index 0)
        # - unknown_value (index 1)
        return len(self.vocab) + self.ns

    def _doc2index(self, document):
        output = [self.word2index[token] if token in self.vocab else self.unknown_value for
                  token in document]
        if self.padding == 'max':
            if len(output) < self.max_length:
                output += [self.padding_value] * (self.max_length - len(output))
            else:
                output = output[:self.max_length]
        return output

    def _update_vocab(self, vocab):
        new_words = sorted(set(vocab) - set(self.vocab))
        self.vocab = self.vocab + new_words
        n = len(self.vocab)
        self.word2index = dict(zip(self.vocab, range(self.ns, n + self.ns)))
        self.index2word = dict(zip(range(self.ns, n + self.ns), self.vocab))

    def _transform(self, raw_documents, is_fit):
        documents = [document.split() for document in raw_documents]
        vocab = sorted(set((reduce(lambda x, y: x + y, documents))))
        max_length_doc = max(len(_) for _ in documents)
        if is_fit:
            self.max_length = max(self.max_length, max_length_doc)
            self._update_vocab(vocab)
        output = [self._doc2index(document) for document in documents]
        return output

    def fit_transform(self, raw_documents):
        return self._transform(raw_documents, is_fit=True)

    def transform(self, raw_documents):
        return self._transform(raw_documents, is_fit=False)



# # Adapted from Yu Zhang's code here: https://github.com/yzhangcs/parser/blob/main/supar/models/dependency.py
# -*- coding: utf-8 -*-
import os
from datetime import datetime

from underthesea.utils import logger
from underthesea.utils.sp_data import Dataset
from underthesea.utils.sp_field import Field
from underthesea.utils.sp_fn import ispunct
from underthesea.utils.sp_init import PRETRAINED
from underthesea.utils.sp_metric import AttachmentMetric
import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from underthesea.transforms.conll import CoNLL
from underthesea.modules.base import CharLSTM, IndependentDropout, BiLSTM, SharedDropout, MLP, Biaffine
from underthesea.modules.bert import BertEmbedding
from underthesea.utils.sp_alg import eisner, mst
import underthesea.modules.nn
from underthesea.utils.util_deep_learning import device


class DependencyParser(underthesea.modules.nn.Model):
    r"""
    The implementation of Biaffine Dependency Parser.

    References:
        - Timothy Dozat and Christopher D. Manning. 2017.
          `Deep Biaffine Attention for Neural Dependency Parsing`_.

    .. _Deep Biaffine Attention for Neural Dependency Parsing:
        https://openreview.net/forum?id=Hk95PK9le
    """
    NAME = 'biaffine-dependency'

    def __init__(
            self,
            n_words=None,
            pad_index=0,
            unk_index=1,
            n_feats=None,
            n_rels=None,
            feat='char',
            n_embed=50,
            n_feat_embed=100,
            n_char_embed=50,
            bert=None,
            n_bert_layers=4,
            embed_dropout=.33,
            max_len=None,
            mix_dropout=.0,
            embeddings=[],
            embed=False,

            n_lstm_hidden=400,
            n_lstm_layers=3,
            lstm_dropout=.33,
            n_mlp_arc=500,
            n_mlp_rel=100,
            mlp_dropout=.33,
            feat_pad_index=0,
            init_pre_train=False,
            transform=None
    ):
        super(DependencyParser, self).__init__()
        self.embed = embed
        self.feat = feat
        self.embeddings = embeddings
        if len(self.embeddings) > 0:
            print(self.embeddings[0])
        self.args = {
            "n_words": n_words,
            'pad_index': pad_index,
            'unk_index': unk_index,
            "n_feats": n_feats,
            "n_rels": n_rels,
            'feat_pad_index': feat_pad_index,
            "feat": feat,
            'tree': False,
            'proj': False,
            'punct': False
        }

        if init_pre_train:
            return

        # the embedding layer
        self.word_embed = nn.Embedding(num_embeddings=n_words,
                                       embedding_dim=n_embed)
        if feat == 'char':
            self.feat_embed = CharLSTM(n_chars=n_feats,
                                       n_embed=n_char_embed,
                                       n_out=n_feat_embed,
                                       pad_index=feat_pad_index)
        elif feat == 'bert':
            self.feat_embed = BertEmbedding(model=bert,
                                            n_layers=n_bert_layers,
                                            n_out=n_feat_embed,
                                            pad_index=feat_pad_index,
                                            max_len=max_len,
                                            dropout=mix_dropout)
            self.n_feat_embed = self.feat_embed.n_out
        elif feat == 'tag':
            self.feat_embed = nn.Embedding(num_embeddings=n_feats,
                                           embedding_dim=n_feat_embed)
        else:
            raise RuntimeError("The feat type should be in ['char', 'bert', 'tag'].")
        self.embed_dropout = IndependentDropout(p=embed_dropout)

        # the lstm layer
        self.lstm = BiLSTM(input_size=n_embed + n_feat_embed,
                           hidden_size=n_lstm_hidden,
                           num_layers=n_lstm_layers,
                           dropout=lstm_dropout)
        self.lstm_dropout = SharedDropout(p=lstm_dropout)

        # the MLP layers
        self.mlp_arc_d = MLP(n_in=n_lstm_hidden * 2,
                             n_out=n_mlp_arc,
                             dropout=mlp_dropout)
        self.mlp_arc_h = MLP(n_in=n_lstm_hidden * 2,
                             n_out=n_mlp_arc,
                             dropout=mlp_dropout)
        self.mlp_rel_d = MLP(n_in=n_lstm_hidden * 2,
                             n_out=n_mlp_rel,
                             dropout=mlp_dropout)
        self.mlp_rel_h = MLP(n_in=n_lstm_hidden * 2,
                             n_out=n_mlp_rel,
                             dropout=mlp_dropout)

        # the Biaffine layers
        self.arc_attn = Biaffine(n_in=n_mlp_arc,
                                 bias_x=True,
                                 bias_y=False)
        self.rel_attn = Biaffine(n_in=n_mlp_rel,
                                 n_out=n_rels,
                                 bias_x=True,
                                 bias_y=True)
        self.criterion = nn.CrossEntropyLoss()
        self.pad_index = pad_index
        self.unk_index = unk_index

        self.transform = transform
        feat = self.feat
        if feat in ('char', 'bert'):
            self.WORD, self.FEAT = self.transform.FORM
        else:
            self.WORD, self.FEAT = self.transform.FORM, self.transform.CPOS
        self.ARC, self.REL = self.transform.HEAD, self.transform.DEPREL
        self.puncts = torch.tensor([i
                                    for s, i in self.WORD.vocab.stoi.items()
                                    if ispunct(s)]).to(device)

    def _get_state_dict(self):
        pass

    @torch.no_grad()
    def predict(
            self,
            data,
            buckets=8,
            batch_size=5000,
            pred=None,
            prob=False,
            tree=True,
            proj=False,
    ):
        r"""
        Args:
            data (list[list] or str):
                The data for prediction, both a list of instances and filename are allowed.
            pred (str):
                If specified, the predicted results will be saved to the file. Default: ``None``.
            buckets (int):
                The number of buckets that sentences are assigned to. Default: 32.
            batch_size (int):
                The number of tokens in each batch. Default: 5000.
            prob (bool):
                If ``True``, outputs the probabilities. Default: ``False``.
            tree (bool):
                If ``True``, ensures to output well-formed trees. Default: ``False``.
            proj (bool):
                If ``True``, ensures to output projective trees. Default: ``False``.
            verbose (bool):
                If ``True``, increases the output verbosity. Default: ``True``.
            kwargs (dict):
                A dict holding the unconsumed arguments that can be used to update the configurations for prediction.

        Returns:
            A :class:`~underthesea.utils.Dataset` object that stores the predicted results.
        """
        self.transform.eval()
        if prob:
            self.transform.append(Field('probs'))

        logger.debug('Loading the data')
        dataset = Dataset(self.transform, data)
        dataset.build(batch_size, buckets)
        logger.debug(f'\n{dataset}')

        logger.debug('Making predictions on the dataset')
        start = datetime.now()
        loader = dataset.loader
        self.eval()

        arcs, rels, probs = [], [], []
        for words, feats in loader:
            mask = words.ne(self.WORD.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            lens = mask.sum(1).tolist()
            s_arc, s_rel = self.forward(words, feats)
            arc_preds, rel_preds = self.decode(s_arc, s_rel, mask, tree, proj)
            arcs.extend(arc_preds[mask].split(lens))
            rels.extend(rel_preds[mask].split(lens))
            if prob:
                arc_probs = s_arc.softmax(-1)
                probs.extend([prob[1:i + 1, :i + 1].cpu() for i, prob in zip(lens, arc_probs.unbind())])
        arcs = [seq.tolist() for seq in arcs]
        rels = [self.REL.vocab[seq.tolist()] for seq in rels]
        preds = {'arcs': arcs, 'rels': rels}
        if prob:
            preds['probs'] = probs

        elapsed = datetime.now() - start

        for name, value in preds.items():
            setattr(dataset, name, value)
        if pred is not None:
            logger.debug(f'Saving predicted results to {pred}')
            self.transform.save(pred, dataset.sentences)
        logger.debug(f'{elapsed}s elapsed, {len(dataset) / elapsed.total_seconds():.2f} Sents/s')

        return dataset

    @torch.no_grad()
    def evaluate(self, loader):
        self.eval()

        total_loss, metric = 0, AttachmentMetric()

        tree = self.args['tree']
        proj = self.args['proj']

        for words, feats, arcs, rels in loader:
            mask = words.ne(self.WORD.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            s_arc, s_rel = self.forward(words, feats)
            loss = self.forward_loss(s_arc, s_rel, arcs, rels, mask)
            arc_preds, rel_preds = self.decode(s_arc, s_rel, mask, tree, proj)
            # ignore all punctuation if not specified
            if not self.args['punct']:
                mask &= words.unsqueeze(-1).ne(self.puncts).all(-1)
            total_loss += loss.item()
            metric(arc_preds, rel_preds, arcs, rels, mask)
        total_loss /= len(loader)

        return total_loss, metric

    @staticmethod
    def _init_model_with_state_dict(state):
        args = state['args']
        transform = state['transform']
        embeddings = state['embeddings']
        model = DependencyParser(
            n_words=args['n_words'],
            n_feats=args['n_feats'],
            n_rels=args['n_feats'],
            pad_index=args['pad_index'],
            unk_index=args['unk_index'],
            # bos_index=args.bos_index,
            feat_pad_index=args['feat_pad_index'],
            transform=transform,
            embeddings=embeddings
        )
        model.load_pretrained(state['pretrained'])
        model.load_state_dict(state['state_dict'], False)
        return model

    @classmethod
    def load(cls, path):
        r"""
        Loads a parser with data fields and pretrained model parameters.

        Args:
            path (str):
                - a string with the shortcut name of a pretrained parser defined in ``underthesea.PRETRAINED``
                  to load from cache or download, e.g., ``'crf-dep-en'``.
                - a path to a directory containing a pre-trained parser, e.g., `./<path>/model`.
            kwargs (dict):
                A dict holding the unconsumed arguments that can be used to update the configurations and initiate the model.

        Examples:
            >>> # from underthesea.models.dependency_parser import DependencyParser
            >>> # parser = DependencyParser.load('vi-dp-v1')
            >>> # parser = DependencyParser.load('./tmp/resources/parsers/dp')
        """
        if os.path.exists(path):
            state = torch.load(path)
        else:
            path = PRETRAINED[path] if path in PRETRAINED else path
            state = torch.hub.load_state_dict_from_url(path)

        model = cls._init_model_with_state_dict(state)
        model.eval()
        model.to(device)
        return model

    def save(self, path):
        model = self
        if hasattr(self, 'module'):
            model = self.module

        state_dict = {k: v.cpu() for k, v in model.state_dict().items()}
        pretrained = state_dict.pop('pretrained.weight', None)
        state = {
            'name': self.NAME,
            'args': self.args,
            'state_dict': state_dict,
            'pretrained': pretrained,
            'transform': self.transform,
            'embeddings': self.embeddings
        }
        torch.save(state, path)

    def load_pretrained(self, embed=None):
        if embed is not None:
            self.pretrained = nn.Embedding.from_pretrained(embed)
            nn.init.zeros_(self.word_embed.weight)
        return self

    def forward(self, words, feats):
        r"""
        Args:
            words (torch.LongTensor): ``[batch_size, seq_len]``.
                Word indices.
            feats (torch.LongTensor):
                Feat indices.
                If feat is ``'char'`` or ``'bert'``, the size of feats should be ``[batch_size, seq_len, fix_len]``.
                if ``'tag'``, the size is ``[batch_size, seq_len]``.

        Returns:
            torch.Tensor, torch.Tensor:
                The first tensor of shape ``[batch_size, seq_len, seq_len]`` holds scores of all possible arcs.
                The second of shape ``[batch_size, seq_len, seq_len, n_labels]`` holds
                scores of all possible labels on each arc.
        """

        batch_size, seq_len = words.shape
        # get the mask and lengths of given batch
        mask = words.ne(self.pad_index)
        ext_words = words
        # set the indices larger than num_embeddings to unk_index
        if hasattr(self, 'pretrained'):
            ext_mask = words.ge(self.word_embed.num_embeddings)
            ext_words = words.masked_fill(ext_mask, self.unk_index)

        # get outputs from embedding layers
        word_embed = self.word_embed(ext_words)
        if hasattr(self, 'pretrained'):
            word_embed += self.pretrained(words)
        feat_embed = self.feat_embed(feats)
        word_embed, feat_embed = self.embed_dropout(word_embed, feat_embed)
        # concatenate the word and feat representations
        embed = torch.cat((word_embed, feat_embed), -1)

        x = pack_padded_sequence(embed, mask.sum(1), True, False)
        x, _ = self.lstm(x)
        x, _ = pad_packed_sequence(x, True, total_length=seq_len)
        x = self.lstm_dropout(x)

        # apply MLPs to the BiLSTM output states
        arc_d = self.mlp_arc_d(x)
        arc_h = self.mlp_arc_h(x)
        rel_d = self.mlp_rel_d(x)
        rel_h = self.mlp_rel_h(x)

        # [batch_size, seq_len, seq_len]
        s_arc = self.arc_attn(arc_d, arc_h)
        # [batch_size, seq_len, seq_len, n_rels]
        s_rel = self.rel_attn(rel_d, rel_h).permute(0, 2, 3, 1)
        # set the scores that exceed the length of each sentence to -inf
        s_arc.masked_fill_(~mask.unsqueeze(1), float('-inf'))

        return s_arc, s_rel

    def forward_loss(self, s_arc, s_rel, arcs, rels, mask):
        r"""
        Args:
            s_arc (~torch.Tensor): ``[batch_size, seq_len, seq_len]``.
                Scores of all possible arcs.
            s_rel (~torch.Tensor): ``[batch_size, seq_len, seq_len, n_labels]``.
                Scores of all possible labels on each arc.
            arcs (~torch.LongTensor): ``[batch_size, seq_len]``.
                The tensor of gold-standard arcs.
            rels (~torch.LongTensor): ``[batch_size, seq_len]``.
                The tensor of gold-standard labels.
            mask (~torch.BoolTensor): ``[batch_size, seq_len]``.
                The mask for covering the unpadded tokens.

        Returns:
            ~torch.Tensor:
                The training loss.
        """

        s_arc, arcs = s_arc[mask], arcs[mask]
        s_rel, rels = s_rel[mask], rels[mask]
        s_rel = s_rel[torch.arange(len(arcs)), arcs]
        arc_loss = self.criterion(s_arc, arcs)
        rel_loss = self.criterion(s_rel, rels)

        return arc_loss + rel_loss

    def decode(self, s_arc, s_rel, mask, tree=False, proj=False):
        r"""
        Args:
            s_arc (~torch.Tensor): ``[batch_size, seq_len, seq_len]``.
                Scores of all possible arcs.
            s_rel (~torch.Tensor): ``[batch_size, seq_len, seq_len, n_labels]``.
                Scores of all possible labels on each arc.
            mask (~torch.BoolTensor): ``[batch_size, seq_len]``.
                The mask for covering the unpadded tokens.
            tree (bool):
                If ``True``, ensures to output well-formed trees. Default: ``False``.
            proj (bool):
                If ``True``, ensures to output projective trees. Default: ``False``.

        Returns:
            ~torch.Tensor, ~torch.Tensor:
                Predicted arcs and labels of shape ``[batch_size, seq_len]``.
        """

        lens = mask.sum(1)
        arc_preds = s_arc.argmax(-1)
        bad = [not CoNLL.istree(seq[1:i + 1], proj)
               for i, seq in zip(lens.tolist(), arc_preds.tolist())]
        if tree and any(bad):
            alg = eisner if proj else mst
            arc_preds[bad] = alg(s_arc[bad], mask[bad])
        rel_preds = s_rel.argmax(-1).gather(-1, arc_preds.unsqueeze(-1)).squeeze(-1)

        return arc_preds, rel_preds



# -*- coding: utf-8 -*-


def classify(X, domain=None, model=None):
    """
    Text classification

    Args:
        X (str): The raw sentence
        domain (str, optional): The domain of the text. Defaults to None.
            Options include:
                - None: general domain
                - 'bank': bank domain
        model (str, optional): The classification model. Defaults to None.
            Options include:
                - None: default underthesea classifier
                - 'prompt': OpenAI prompt model

    Returns:
        list: A list containing the categories of the sentence
    """
    if X == "":
        return None

    if model == 'prompt':
        from underthesea.pipeline.classification import classification_prompt
        args = {
            "domain": domain
        }
        return classification_prompt.classify(X, **args)

    if domain == 'bank':
        from underthesea.pipeline.classification import bank
        return bank.classify(X)

    from underthesea.pipeline.classification import vntc
    return vntc.classify(X)



import pickle
from os.path import join, dirname

from underthesea.util.singleton import Singleton


@Singleton
class Dictionary:
    def __init__(self):
        """ contains 13k words in this version
        """
        filepath = join(dirname(__file__), "dictionary.data")
        with open(filepath, "rb") as f:
            self.words = pickle.load(f)

    def lookup(self, text):
        if text in self.words:
            return self.words[text]
        return None



import shutil
from os.path import isdir

from underthesea.corpus import Corpus
from underthesea.data_fetcher import DataFetcher
from underthesea.file_utils import DATASETS_FOLDER


class VLSP2020_DP(Corpus):
    def __init__(self):
        VLSP2020_DP_FOLDER = f'{DATASETS_FOLDER}/VLSP2020-DP'
        train_file = f'{VLSP2020_DP_FOLDER}/train.txt'
        test_file = f'{VLSP2020_DP_FOLDER}/test.txt'
        self._train = train_file
        self._test = test_file
        self._dev = test_file


class VLSP2020_DP_SAMPLE(Corpus):
    name = "VLSP2020_DP_SAMPLE"
    folder = f'{DATASETS_FOLDER}/{name}'
    REPO_DATA = {
        "url": 'https://github.com/undertheseanlp/resources/releases/download/1.3.x/VLSP2020_DP_SAMPLE-1.0-alpha.zip',
        "url_filename": "VLSP2020_DP_SAMPLE-1.0-alpha.zip",
        "cache_dir": f'datasets/{name}',
    }

    @staticmethod
    def download_file():
        if isdir(VLSP2020_DP_SAMPLE.folder):
            return
        shutil.rmtree(VLSP2020_DP_SAMPLE.folder, ignore_errors=True)
        DataFetcher.download_zip_file_to_cache(VLSP2020_DP_SAMPLE.REPO_DATA)

    def __init__(self):
        self.download_file()

        train_file = f'{self.folder}/train.txt'
        dev_file = f'{self.folder}/dev.txt'
        test_file = f'{self.folder}/test.txt'
        self._train = train_file
        self._test = dev_file
        self._dev = test_file



import os
import pickle
from functools import partial
from typing import Deque

import fire
import jax
import jax.numpy as jnp
import jax.tools.colab_tpu
import matplotlib.pyplot as plt
import optax
from tqdm.auto import tqdm

from .acoustic_trainer import initial_state, loss_vag, val_loss_fn
from .config import FLAGS
from .data_loader import load_textgrid_wav
from .dsp import MelFilter
from .utils import print_flags


def setup_colab_tpu():
    jax.tools.colab_tpu.setup_tpu()


def train(
    batch_size: int = 32,
    steps_per_update: int = 10,
    learning_rate: float = 1024e-6,
):
    """Train acoustic model on multiple cores (TPU)."""
    lr_schedule = optax.exponential_decay(learning_rate, 50_000, 0.5, staircase=True)

    optimizer = optax.chain(
        optax.clip_by_global_norm(1.0),
        optax.adamw(lr_schedule, weight_decay=FLAGS.weight_decay),
    )

    def update_step(prev_state, inputs):
        params, aux, rng, optim_state = prev_state
        rng, new_rng = jax.random.split(rng)
        (loss, new_aux), grads = loss_vag(params, aux, rng, inputs)
        grads = jax.lax.pmean(grads, axis_name="i")
        updates, new_optim_state = optimizer.update(grads, optim_state, params)
        new_params = optax.apply_updates(params, updates)
        next_state = (new_params, new_aux, new_rng, new_optim_state)
        return next_state, loss

    @partial(jax.pmap, axis_name="i")
    def update(params, aux, rng, optim_state, inputs):
        states, losses = jax.lax.scan(
            update_step, (params, aux, rng, optim_state), inputs
        )
        return states, jnp.mean(losses)

    print(jax.devices())
    num_devices = jax.device_count()
    train_data_iter = load_textgrid_wav(
        FLAGS.data_dir,
        FLAGS.max_phoneme_seq_len,
        batch_size * num_devices * steps_per_update,
        FLAGS.max_wave_len,
        "train",
    )
    val_data_iter = load_textgrid_wav(
        FLAGS.data_dir,
        FLAGS.max_phoneme_seq_len,
        batch_size,
        FLAGS.max_wave_len,
        "val",
    )
    melfilter = MelFilter(
        FLAGS.sample_rate,
        FLAGS.n_fft,
        FLAGS.mel_dim,
        FLAGS.fmin,
        FLAGS.fmax,
    )
    batch = next(train_data_iter)
    batch = jax.tree_map(lambda x: x[:1], batch)
    batch = batch._replace(mels=melfilter(batch.wavs.astype(jnp.float32) / (2 ** 15)))
    params, aux, rng, optim_state = initial_state(optimizer, batch)
    losses = Deque(maxlen=1000)
    val_losses = Deque(maxlen=100)

    last_step = -steps_per_update

    # loading latest checkpoint
    ckpt_fn = FLAGS.ckpt_dir / "acoustic_latest_ckpt.pickle"
    if ckpt_fn.exists():
        print("Resuming from latest checkpoint at", ckpt_fn)
        with open(ckpt_fn, "rb") as f:
            dic = pickle.load(f)
            last_step, params, aux, rng, optim_state = (
                dic["step"],
                dic["params"],
                dic["aux"],
                dic["rng"],
                dic["optim_state"],
            )

    tr = tqdm(
        range(
            last_step + steps_per_update, FLAGS.num_training_steps + 1, steps_per_update
        ),
        desc="training",
        total=FLAGS.num_training_steps // steps_per_update + 1,
        initial=last_step // steps_per_update + 1,
    )

    params, aux, rng, optim_state = jax.device_put_replicated(
        (params, aux, rng, optim_state), jax.devices()
    )

    def batch_reshape(batch):
        return jax.tree_map(
            lambda x: jnp.reshape(x, (num_devices, steps_per_update, -1) + x.shape[1:]),
            batch,
        )

    for step in tr:
        batch = next(train_data_iter)
        batch = batch_reshape(batch)
        (params, aux, rng, optim_state), loss = update(
            params, aux, rng, optim_state, batch
        )
        losses.append(loss)

        if step % 10 == 0:
            val_batch = next(val_data_iter)
            val_loss, val_aux, predicted_mel, gt_mel = val_loss_fn(
                *jax.tree_map(lambda x: x[0], (params, aux, rng)), val_batch
            )
            val_losses.append(val_loss)
            attn = jax.device_get(val_aux["acoustic_model"]["attn"])
            predicted_mel = jax.device_get(predicted_mel[0])
            gt_mel = jax.device_get(gt_mel[0])

        if step % 1000 == 0:
            loss = jnp.mean(sum(losses)).item() / len(losses)
            val_loss = sum(val_losses).item() / len(val_losses)
            tr.write(f"step {step}  train loss {loss:.3f}  val loss {val_loss:.3f}")

            # saving predicted mels
            plt.figure(figsize=(10, 10))
            plt.subplot(3, 1, 1)
            plt.imshow(predicted_mel.T, origin="lower", aspect="auto")
            plt.subplot(3, 1, 2)
            plt.imshow(gt_mel.T, origin="lower", aspect="auto")
            plt.subplot(3, 1, 3)
            plt.imshow(attn.T, origin="lower", aspect="auto")
            plt.tight_layout()
            plt.savefig(FLAGS.ckpt_dir / f"mel_{step}.png")
            plt.close()

            # saving checkpoint
            with open(ckpt_fn, "wb") as f:
                params_, aux_, rng_, optim_state_ = jax.tree_map(
                    lambda x: x[0], (params, aux, rng, optim_state)
                )
                pickle.dump(
                    {
                        "step": step,
                        "params": params_,
                        "aux": aux_,
                        "rng": rng_,
                        "optim_state": optim_state_,
                    },
                    f,
                )


if __name__ == "__main__":
    # we don't use these flags.
    del FLAGS.batch_size
    del FLAGS.learning_rate
    del FLAGS.duration_learning_rate
    del FLAGS.duration_lstm_dim
    del FLAGS.duration_embed_dropout_rate

    print_flags(FLAGS.__dict__)

    if "COLAB_TPU_ADDR" in os.environ:
        setup_colab_tpu()

    if not FLAGS.ckpt_dir.exists():
        print("Create checkpoint dir at", FLAGS.ckpt_dir)
        FLAGS.ckpt_dir.mkdir(parents=True, exist_ok=True)

    fire.Fire(train)



import logging
import os
import shutil
from os.path import join

import pycrfsuite
from seqeval.metrics import classification_report
from underthesea_core import CRFFeaturizer

logger = logging.getLogger(__name__)
logger.setLevel(10)
FORMAT = "%(asctime)-15s %(message)s"
logging.basicConfig(format=FORMAT)


class CRFTrainer:
    def __init__(self, model, training_args, train_dataset=None, test_dataset=None):
        self.model = model
        self.training_args = training_args
        self.output_dir = training_args["output_dir"]
        self.model_params = training_args["params"]
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset

    def train(self):
        # create output_dir directory
        output_dir = self.output_dir
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.makedirs(output_dir)
        features = self.model.features
        print(features)

        featurizer = CRFFeaturizer(features, set())
        logger.info("Start feature extraction")
        # featurizer = TaggedTransformer(features)
        # X_train, y_train = featurizer.transform(self.train_dataset, contain_labels=True)
        # X_test, y_test = featurizer.transform(self.test_dataset, contain_labels=True)

        X_train = featurizer.process(self.train_dataset)
        y_train = []
        for s in self.train_dataset:
            yi = [t[-1] for t in s]
            y_train.append(yi)
        X_test = featurizer.process(self.test_dataset)
        y_test = []
        for s in self.test_dataset:
            yi = [t[-1] for t in s]
            y_test.append(yi)
        logger.info("Finish feature extraction")

        # Train
        logger.info("Start train")
        trainer = pycrfsuite.Trainer(verbose=True)
        for xseq, yseq in zip(X_train, y_train):
            trainer.append(xseq, yseq)
        trainer.set_params(self.model_params)
        model_path = join(output_dir, "models.bin")
        trainer.train(model_path)
        logger.info("Finish train")
        self.model.save(output_dir)

        # Tagger
        logger.info("Start tagger")
        tagger = pycrfsuite.Tagger()
        tagger.open(model_path)
        y_pred = [tagger.tag(x_seq) for x_seq in X_test]
        sentences = [[item[0] for item in sentence] for sentence in self.test_dataset]
        sentences = zip(sentences, y_test, y_pred)
        texts = []
        for s in sentences:
            tokens, y_true, y_pred_ = s
            tokens_ = []
            for i in range(len(tokens)):
                if tokens[i] == "":
                    token = "X"
                else:
                    token = tokens[i]
                tokens_.append(token + "\t" + y_true[i] + "\t" + y_pred_[i])
            text = "\n".join(tokens_)
            texts.append(text)
        text = "\n\n".join(texts)
        tmp_output_path = join(output_dir, "test_output.txt")
        open(tmp_output_path, "w").write(text)
        print("Classification report:\n")
        print(classification_report(y_test, y_pred, digits=3))
        logger.info("Finish tagger")



from os.path import join, dirname
from underthesea.util.singleton import Singleton
import pycrfsuite
from .tagged_feature import word2features


# [Deprecating] This file will be deprecated in the version 6.5.0

@Singleton
class CRFPOSTagPredictor:
    def __init__(self):
        self.model = pycrfsuite.Tagger()
        filepath = join(dirname(__file__), "pos_crf_2017_10_11.bin")
        self.model.open(filepath)

        template = [
            "T[-2].lower", "T[-1].lower", "T[0].lower", "T[1].lower",
            "T[2].lower",
            "T[0].istitle", "T[-1].istitle", "T[1].istitle",
            # word unigram and bigram
            "T[-2]", "T[-1]", "T[0]", "T[1]", "T[2]",
            "T[-2,-1]", "T[-1,0]", "T[0,1]", "T[1,2]",
            # pos unigram and bigram
            "T[-3][1]", "T[-2][1]", "T[-1][1]",
            "T[-3,-2][1]", "T[-2,-1][1]",
        ]
        self.transformer = TaggedTransformer(template)

    def predict(self, sentence, format=None):
        tokens = [(token, "X") for token in sentence]
        x = self.transformer.transform([tokens])[0][0]
        tags = self.model.tag(x)
        return list(zip(sentence, tags))


class TaggedTransformer:
    def __init__(self, template=None):
        self.template = template

    def transform(self, sentences):
        X = [self.sentence2features(s) for s in sentences]
        y = [self.sentence2labels(s) for s in sentences]
        return X, y

    def sentence2features(self, s):
        return [word2features(s, i, self.template) for i in range(len(s))]

    def sentence2labels(self, s):
        return [row[-1] for row in s]



from enum import Enum
from os.path import join
import json
import joblib
from sklearn.preprocessing import MultiLabelBinarizer
import warnings

from underthesea.corpus.data import Sentence, Label

warnings.filterwarnings("ignore", ".*", UserWarning)


class Model:
    pass


class TEXT_CLASSIFIER_ESTIMATOR(Enum):
    FAST_TEXT = "FAST_TEXT"
    SVC = "SVC"
    PIPELINE = "PIPELINE"


class TextClassifier(Model):

    def __init__(self, estimator: TEXT_CLASSIFIER_ESTIMATOR, multilabel=False, **params):
        self.multilabel = multilabel
        if multilabel:
            self.y_encoder = MultiLabelBinarizer()
        self.estimator = estimator
        self.params = params
        if estimator == TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT:
            self.ft = None
        if estimator == TEXT_CLASSIFIER_ESTIMATOR.SVC:
            self.svc = None
        if estimator == TEXT_CLASSIFIER_ESTIMATOR.PIPELINE:
            if "pipeline" in params:
                self.pipeline = params["pipeline"]
            else:
                self.pipeline = None

    @staticmethod
    def load(model_folder):
        with open(join(model_folder, "metadata.json")) as f:
            metadata = json.loads(f.read())
        if metadata['estimator'] == 'SVC':
            estimator = TEXT_CLASSIFIER_ESTIMATOR.SVC
        if metadata['estimator'] == 'FAST_TEXT':
            estimator = TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT
        if metadata['estimator'] == 'PIPELINE':
            estimator = TEXT_CLASSIFIER_ESTIMATOR.PIPELINE

        # GH-304: remove fasttext
        # if estimator == TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT:
        #     model_file = join(model_folder, "model.bin")
        #     classifier = TextClassifier(estimator=TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT)
        #     classifier.ft = fastText.load_model(model_file)
        #     return classifier

        if estimator == TEXT_CLASSIFIER_ESTIMATOR.SVC:
            classifier = TextClassifier(estimator=TEXT_CLASSIFIER_ESTIMATOR.SVC)
            classifier.svc = joblib.load(join(model_folder, "estimator.joblib"))
            x_transformer = joblib.load(join(model_folder, "x_transformer.joblib"))
            classifier.x_transformer = x_transformer
            y_transformer = joblib.load(join(model_folder, "y_transformer.joblib"))
            classifier.y_transformer = y_transformer
            return classifier

        if estimator == TEXT_CLASSIFIER_ESTIMATOR.PIPELINE:
            classifier = TextClassifier(estimator=TEXT_CLASSIFIER_ESTIMATOR.PIPELINE)
            if "multilabel" in metadata:
                if metadata["multilabel"]:
                    classifier.multilabel = True
                    classifier.y_encoder = joblib.load(join(model_folder, "y_encoder.joblib"))
            classifier.pipeline = joblib.load(join(model_folder, "pipeline.joblib"))

            return classifier

    def predict(self, sentence: Sentence):
        if self.estimator == TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT:
            values, scores = self.ft.predict(sentence.text)
            labels = []
            for value, score in zip(values, scores):
                value = value.replace("__label__", "")
                label = Label(value, score)
                labels.append(label)
            sentence.add_labels(labels)

        if self.estimator == TEXT_CLASSIFIER_ESTIMATOR.SVC:
            text = sentence.text
            X = self.x_transformer.transform([text])
            y = self.svc.predict(X)
            y = self.y_transformer.inverse_transform(y)
            sentence.add_labels(y)

        if self.estimator == TEXT_CLASSIFIER_ESTIMATOR.PIPELINE:
            text = sentence.text
            y = self.pipeline.predict([text])
            if self.multilabel:
                y = self.y_encoder.inverse_transform(y)
                y = list(y[0])
            else:
                y = list(y)
            sentence.add_labels(y)



# -*- coding: utf-8 -*-

import os
from random import Random

import torch
import torch.distributed as dist
import torch.nn as nn


class DistributedDataParallel(nn.parallel.DistributedDataParallel):

    def __init__(self, module, **kwargs):
        super().__init__(module, **kwargs)

    def __getattr__(self, name):
        wrapped = super().__getattr__('module')
        if hasattr(wrapped, name):
            return getattr(wrapped, name)
        return super().__getattr__(name)


def init_device(device, backend='nccl', host=None, port=None):
    os.environ['CUDA_VISIBLE_DEVICES'] = device
    if torch.cuda.device_count() > 1:
        host = host or os.environ.get('MASTER_ADDR', 'localhost')
        port = port or os.environ.get('MASTER_PORT', str(Random(0).randint(10000, 20000)))
        os.environ['MASTER_ADDR'] = host
        os.environ['MASTER_PORT'] = port
        dist.init_process_group(backend)
        torch.cuda.set_device(dist.get_rank())


def is_master():
    return not dist.is_initialized() or dist.get_rank() == 0



from os.path import join
import re
from underthesea.data_fetcher import DataFetcher
from underthesea.file_utils import DATASETS_FOLDER

_CITATION = None

_DESCRIPTION = None

_HOMEPAGE = None

_LICENSE = None


class LabelEncoder:
    def __init__(self):
        self.index2label = {}
        self.label2index = {}
        self.vocab_size = 0

    def encode(self, labels):
        if type(labels) is list:
            return [self.encode(label) for label in labels]
        label = labels  # label is a string
        if label not in self.label2index:
            index = self.vocab_size
            self.label2index[label] = index
            self.index2label[index] = label
            self.vocab_size += 1
            return index
        else:
            return self.label2index[label]


class UITABSAHotel:
    NAME = "UIT_ABSA_HOTEL"
    VERSION = "1.0"

    def __init__(self, training="aspect"):
        DataFetcher.download_data(UITABSAHotel.NAME, None)
        train_file = join(DATASETS_FOLDER, UITABSAHotel.NAME, "Train_Hotel.txt")
        dev_file = join(DATASETS_FOLDER, UITABSAHotel.NAME, "Dev_Hotel.txt")
        test_file = join(DATASETS_FOLDER, UITABSAHotel.NAME, "Test_Hotel.txt")
        print("Currently training: %s (aspect or polarity)" % training)

        self.training = training  # aspect or polarity
        self.label_encoder = LabelEncoder()
        self.train = self._extract_sentences(train_file)
        self.dev = self._extract_sentences(dev_file)
        self.test = self._extract_sentences(test_file)
        self.num_labels = self.label_encoder.vocab_size

    def _join_labels(self, label):
        return "#".join(label)

    def _extract_sentence(self, sentence):
        sentence_id, text, labels = sentence.split("\n")
        labels = re.findall("\{(?P<aspect>.*?), (?P<polarity>.*?)\}", labels)
        aspect_labels = [label[0] for label in labels]
        polarity_labels = [label[1] for label in labels]  # some odd polarity labels outside of pos, neg, neutral

        if self.training == "aspect":
            label_ids = self.label_encoder.encode(aspect_labels)
        else:
            label_ids = self.label_encoder.encode(polarity_labels)

        return {
            "sentence_id": sentence_id,
            "text": text,
            "labels": labels,
            "aspect_labels": aspect_labels,
            "polarity_labels": polarity_labels,
            "label_ids": label_ids
        }

    def _extract_sentences(self, file):
        with open(file, encoding='utf-8') as f:
            content = f.read()
            sentences = content.split("\n\n")
            sentences = [self._extract_sentence(s) for s in sentences]
        return sentences


if __name__ == '__main__':
    dataset = UITABSAHotel(training="aspect")



from pathlib import Path
from typing import List, Union
from underthesea.corpus import Corpus
from underthesea.corpus.data import Sentence


class CategorizedCorpus(Corpus):
    def __init__(
        self,
        train: List[Sentence],
        dev: List[Sentence],
        test: List[Sentence]
    ):
        self._train: List[Sentence] = train
        self._dev: List[Sentence] = dev
        self._test: List[Sentence] = test

    @property
    def train(self) -> List[Sentence]:
        return self._train

    @property
    def dev(self) -> List[Sentence]:
        return self._dev

    @property
    def test(self) -> List[Sentence]:
        return self._test

    def __str__(self) -> str:
        return "CategorizedCorpus: %d train + %d dev + %d test sentences" % (
            len(self.train),
            len(self.dev),
            len(self.test),
        )

    def save(self, data_folder: str):
        self.save_sentences(self._train, f"{data_folder}/train.txt")
        self.save_sentences(self._dev, f"{data_folder}/dev.txt")
        self.save_sentences(self._test, f"{data_folder}/test.txt")

    def save_sentences(self, sentences: List[Sentence], path_to_file: Union[Path, str]):
        with open(path_to_file, "w") as f:
            for sentence in sentences:
                f.write(sentence.to_text_classification_format() + "\n")



import math
from underthesea.dictionary import Dictionary
from collections import Counter, defaultdict
from underthesea.utils import logger
from datetime import date, timedelta
import pandas as pd


def get_words_pos(dictionary):
    output = {}
    for word in dictionary.words:
        data = dictionary.words[word]
        pos = list(set([_['pos'] for _ in data]))
        output[word] = pos
    return output


dictionary = Dictionary.Instance()
n_words = len(dictionary.words)
print(f"Load dictionary with {n_words} words.")
words_pos = get_words_pos(dictionary)


def computeIDF(docCounters):
    """Get Inverse Document Frequency value for each word"""
    n = len(docCounters)
    idfDict = defaultdict(int)
    for doc in docCounters:
        for word, freq in doc.items():
            if freq > 0:
                idfDict[word] += 1

    for word, val in idfDict.items():
        idfDict[word] = math.log10(n / float(val))
    return idfDict


class UDAnalyzer:
    def __init__(self):
        self.total_errors = 0

    def _get_words(self, sentences):
        tags = [s.rows for s in sentences]
        tags = [t for sublist in tags for t in sublist]
        words = [t[0].lower() for t in tags]
        return words

    def _get_doc_sents(self, dataset):
        """Return doc as key and doc sentence array as value"""
        doc_sents = defaultdict(list)
        for s in dataset:
            doc = s.headers['doc_url']
            doc_sents[doc].append(s)
        return doc_sents

    def get_doc_word_counters(self, dataset):
        """Get word count by doc url"""
        data = self._get_doc_sents(dataset)
        doc_word_counters = {}
        for doc, sents in data.items():
            tags = [s.rows for s in sents]
            tags = [t for sublist in tags for t in sublist]
            words = [t[0].lower() for t in tags]
            doc_word_counters[doc] = Counter(words)
        return doc_word_counters

    def analyze_words_pos(self, sentences):
        tags = [s.rows for s in sentences]
        tags = [t for sublist in tags for t in sublist]
        df = pd.DataFrame(tags, columns=["word", "pos", "order", "dep"])
        return df

    def analyze_words(self, sentences):
        tags = [s.rows for s in sentences]
        tags = [t for sublist in tags for t in sublist]
        words = [t[0].lower() for t in tags if t[1] not in ["Np", "CH"]]
        noun_phrases = [t[0] for t in tags if t[1] == "Np"]
        counter = Counter(words)
        corpus_words = set(counter.keys())
        dictionary_words = set(dictionary.words)
        oov = corpus_words - dictionary_words

        print("Most common words:\n", counter.most_common(50))
        print("Corpus words: ", len(corpus_words))
        print("Noun phrases:\n", set(noun_phrases))
        print("Short OOV words:\n", sorted([item for item in oov if len(item.split(" ")) < 3]))
        print("Long OOV words:\n", sorted([item for item in oov if len(item.split(" ")) >= 3]))
        return counter

    def analyze_all_words(self, dataset):
        return self.analyze_words(dataset)

    def get_today_sentences(self, dataset):
        yesterday = (date.today() - timedelta(days=1)).strftime('%Y%m%d')
        sentences = [s for s in dataset if s.headers['date'] == yesterday]
        return sentences

    def analyze_today_words(self, dataset):
        sentences = self.get_today_sentences(dataset)
        return self.analyze_words(sentences)

    def analyze_sent_ids(self, dataset):
        """Get sent_id of all sentences"""
        sent_ids = [s.headers['sent_id'] for s in dataset]
        counter = Counter(sent_ids)

        duplicate_ids = [key for key in counter if counter[key] > 1]
        if len(duplicate_ids) > 0:
            logger.error('[ERROR] duplicate_ids' + str(duplicate_ids))
        else:
            logger.debug("sent_ids is valid.")

        missing_ids = [str(i + 1) for i in range(len(sent_ids)) if str(i + 1) not in sent_ids]
        if len(missing_ids) > 0:
            logger.error('[ERROR] missing_ids' + str(missing_ids))
        return sent_ids

    def analyze_doc_sent_freq(self, dataset):
        """Get sentence count by doc url"""
        data = self._get_doc_sents(dataset)
        print("Number of doc URLs %s" % len(data))
        doc_sent_counts = [(doc, len(sents)) for doc, sents in data.items()]
        return doc_sent_counts

    def analyze_dataset_len(self, dataset):
        print("Number of sentences", len(dataset))

    def analyze(self, dataset):
        self.analyze_dataset_len(dataset)
        self.analyze_sent_ids(dataset)
        self.analyze_words(dataset)
        self.analyze_doc_sent_freq(dataset)
        self.get_doc_word_counters(dataset)



from os.path import join


class TaggedCorpus:
    def __init__(self, train, test):
        self.train = train
        self.test = test

    def downsample(self, percentage):
        n = int(len(self.train) * percentage)
        self.train = self.train[:n]
        n = int(len(self.test) * percentage)
        self.test = self.test[:n]
        return self


class DataReader:
    @staticmethod
    def load_tagged_corpus(data_folder, train_file=None, test_file=None):
        train = DataReader.__read_tagged_data(join(data_folder, train_file))
        test = DataReader.__read_tagged_data(join(data_folder, test_file))
        tagged_corpus = TaggedCorpus(train, test)
        return tagged_corpus

    @staticmethod
    def __read_tagged_data(data_file):
        sentences = []
        raw_sentences = open(data_file).read().strip().split("\n\n")
        for s in raw_sentences:
            tagged_sentence = [node.split("\t") for node in s.split("\n") if not node.startswith("#")]
            sentences.append(tagged_sentence)
        return sentences



from os.path import join
from pathlib import Path

import joblib
import pycrfsuite
from underthesea.transformer.tagged import TaggedTransformer


class PyCRFSequenceTagger:
    def __init__(self, features=None, dictionary=set()):
        self.features = features
        self.dictionary = dictionary
        self.crf_tagger = None
        self.path_model = "models.bin"
        self.path_features = "features.bin"
        self.path_dictionary = "dictionary.bin"

    def save(self, base_path):
        joblib.dump(self.features, join(base_path, self.path_features))

    def load(self, base_path):
        model_path = str(Path(base_path) / self.path_model)
        crf_tagger = pycrfsuite.Tagger()
        crf_tagger.open(model_path)
        features = joblib.load(join(base_path, self.path_features))
        featurizer = TaggedTransformer(features)
        self.featurizer = featurizer
        self.crf_tagger = crf_tagger

    def predict(self, tokens):
        tokens = [[token] for token in tokens]
        X = self.featurizer.transform([tokens])[0]
        tags = self.crf_tagger.tag(X)
        return tags



from typing import List, Union


class Label:
    def __init__(self, value: str, score: float = 1.0):
        self.value = value
        self.score = score

    @property
    def value(self):
        return self._value

    @property
    def score(self):
        return self._score

    @value.setter
    def value(self, value):
        self._value = value

    @score.setter
    def score(self, score):
        if 0.0 <= score <= 1.0:
            self._score = score
        else:
            self._score = 1.0

    def __str__(self):
        return "{} ({})".format(self._value, self._score)

    def __repr__(self):
        return "{} ({})".format(self._value, self._score)


class Sentence:
    def __init__(
        self,
        text: str = None,
        labels: Union[List[Label], List[str]] = None
    ):
        self.text = text
        self.labels = labels

    def __str__(self) -> str:
        return f'Sentence: "{self.text}" - Labels: {self.labels}'

    def __repr__(self) -> str:
        return f'Sentence: "{self.text}" - Labels: {self.labels}'

    def to_text_classification_format(self) -> str:
        labels_text = " ".join([f"__label__{label.value}" for label in self.labels])
        output = f"{labels_text} {self.text}"
        return output

    def add_labels(self, labels: Union[List[Label], List[str]]):
        for label in labels:
            if type(label) is str:
                label = Label(label)
            if not self.labels:
                self.labels = []
            self.labels.append(label)


class Document:
    def __init__(self, id):
        """
        :param id id of document
        :type id: str
        """
        self.id = id
        self.content = None
        self.sentences = None

    def set_content(self, content):
        self.content = content

    def set_sentences(self, sentences):
        self.sentences = sentences



import os
import shutil
import zipfile
from enum import Enum
from os.path import dirname, join
from pathlib import Path
import yaml
from underthesea.file_utils import cached_path, UNDERTHESEA_FOLDER
from underthesea.utils import print_table

MISS_URL_ERROR = "Caution:\n  With closed license model, you must provide URL to download"

CD = dirname(__file__)
with open(join(CD, "models.yaml")) as f:
    REPO = yaml.safe_load(f)


class UTSModel(Enum):
    tc_general = "TC_GENERAL"
    tc_bank = "TC_BANK"
    tc_general_v131 = "TC_GENERAL_V131"
    tc_bank_v131 = "TC_BANK_V131"
    sa_general = "SA_GENERAL"
    sa_bank = "SA_BANK"
    sa_bank_v131 = "SA_BANK_V131"
    lang_detect_fast_text = "LANG_DETECT_FAST_TEXT"

# flake8: noqa: C901


class ModelFetcher:

    @staticmethod
    def download_zip(model_info):
        cache_dir = Path(UNDERTHESEA_FOLDER) / model_info["cache_dir"] / model_info["model_path"]
        cached_path(model_info["url"], cache_dir=cache_dir)
        zip_filepath = cache_dir / model_info["filename"]
        with zipfile.ZipFile(zip_filepath) as zip_file:
            zip_file.extractall(cache_dir)
        os.remove(zip_filepath)

    @staticmethod
    def download(model_name):
        if model_name not in REPO:
            print(f"No matching distribution found for '{model_name}'")
            return

        model_path = REPO[model_name]["model_path"]
        cache_dir = REPO[model_name]["cache_dir"]
        model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / model_path
        if Path(model_path).exists():
            print(f"Model is already existed: '{model_name}' in {model_path}")
            return

        if model_name == "TC_GENERAL":
            url = "https://www.dropbox.com/s/866offu8wglrcej/tc_svm_vntc_20190607.zip?dl=1"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_vntc_20190607.zip?dl=1"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_vntc_20190607",
                Path(UNDERTHESEA_FOLDER) / cache_dir / "TC_GENERAL",
            )
            os.remove(model_path)

        if model_name == "TC_GENERAL_V131":
            url = "https://github.com/undertheseanlp/playground/releases/download/1.3.x/tc_svm_vntc_20201228.zip"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_vntc_20201228.zip"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_vntc_20201228",
                Path(UNDERTHESEA_FOLDER) / cache_dir / "TC_GENERAL_V131",
            )
            os.remove(model_path)

        if model_name == "TC_BANK_V131":
            model_id = "tc_svm_ubc-1_20210107"
            url = f"https://github.com/undertheseanlp/playground/releases/download/1.3.x/{model_id}.zip"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / f"{model_id}.zip"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_id,
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_name,
            )
            os.remove(model_path)

        if model_name == "TC_BANK":
            url = "https://www.dropbox.com/s/prrjlypbrr6ze6p/tc_svm_uts2017_bank_20190607.zip?dl=1"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_uts2017_bank_20190607.zip?dl=1"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / "tc_svm_uts2017_bank_20190607",
                Path(UNDERTHESEA_FOLDER) / cache_dir / "TC_BANK",
            )
            os.remove(model_path)

        if model_name == "SA_GENERAL":
            url = "https://www.dropbox.com/s/xfj1ity3egabv77/sa_svm_aivivn2019_20190615.zip?dl=1"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / "sa_svm_aivivn2019_20190615.zip?dl=1"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / "sa_svm_aivivn2019_20190615",
                Path(UNDERTHESEA_FOLDER) / cache_dir / "SA_GENERAL",
            )
            os.remove(model_path)

        if model_name == "SA_GENERAL_V131":
            model_id = "sa_svm_vlsp2016-sa_20210107"
            url = f"https://github.com/undertheseanlp/playground/releases/download/1.3.x/{model_id}.zip"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / f"{model_id}.zip"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_id,
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_name,
            )
            os.remove(model_path)

        if model_name == "SA_BANK":
            url = "https://www.dropbox.com/s/yo6sf6ofpdb3hlh/sa_svm_uts2017_bank_20190611.zip?dl=1"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / "sa_svm_uts2017_bank_20190611.zip?dl=1"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / "sa_svm_uts2017_bank_20190611",
                Path(UNDERTHESEA_FOLDER) / cache_dir / "SA_BANK",
            )
            os.remove(model_path)

        if model_name == "SA_BANK_V131":
            model_id = "sa_svm_ubs-1_20210107"
            url = f"https://github.com/undertheseanlp/playground/releases/download/1.3.x/{model_id}.zip"
            cached_path(url, cache_dir=cache_dir)
            model_path = Path(UNDERTHESEA_FOLDER) / cache_dir / f"{model_id}.zip"
            cache_folder = Path(UNDERTHESEA_FOLDER) / cache_dir
            with zipfile.ZipFile(model_path) as zip_file:
                zip_file.extractall(cache_folder)
            os.rename(
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_id,
                Path(UNDERTHESEA_FOLDER) / cache_dir / model_name,
            )
            os.remove(model_path)

        if model_name == "VIET_TTS_V0_4_1":
            ModelFetcher.download_zip(REPO[model_name])

        if model_name == "LANG_DETECT_FAST_TEXT":
            url = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"
            cached_path(url, cache_dir=cache_dir)

    @staticmethod
    def list(all=False):
        models = []
        for key in REPO:
            name = key
            type = REPO[key]["type"]
            license = REPO[key]["license"]
            year = REPO[key]["year"]
            directory = Path(REPO[key]["cache_dir"]) / REPO[key]["model_path"]
            if not all:
                if license == "Close":
                    continue
            if license == "Close":
                license = "Close*"
            models.append([name, type, license, year, directory])

        print_table(models, headers=["Name", "Type", "License", "Year", "Directory"])

        if all:
            print(f"\n{MISS_URL_ERROR}")

    @staticmethod
    def remove(model_name):
        if model_name not in REPO:
            print(f"No matching distribution found for '{model_name}'")
            return
        model = REPO[model_name]
        cache_dir = Path(UNDERTHESEA_FOLDER) / model["cache_dir"] / model["model_path"]
        if cache_dir.is_dir():
            shutil.rmtree(cache_dir)
        print(f"Model {model_name} is removed.")

    @staticmethod
    def get_model_path(model):
        if model == UTSModel.tc_bank:
            return Path(UNDERTHESEA_FOLDER) / "models" / "TC_BANK"

        if model == UTSModel.tc_general:
            return Path(UNDERTHESEA_FOLDER) / "models" / "TC_GENERAL"

        if model == UTSModel.sa_general:
            return Path(UNDERTHESEA_FOLDER) / "models" / "SA_GENERAL"

        if model == UTSModel.sa_bank:
            return Path(UNDERTHESEA_FOLDER) / "models" / "SA_BANK"

        if model == UTSModel.sa_bank:
            return Path(UNDERTHESEA_FOLDER) / "models" / "SA_BANK"

        if model == "LANG_DETECT_FAST_TEXT":
            return Path(UNDERTHESEA_FOLDER) / "models" / "lid.176.bin"
        
        return Path(UNDERTHESEA_FOLDER) / "models" / model


if __name__ == '__main__':
    ModelFetcher.list()
    ModelFetcher.download("VIET_TTS_V0_4_1")



from underthesea.data_fetcher import DataFetcher
from unittest import TestCase


class TestDataFetcher(TestCase):
    def test_1(self):
        data = "CP_Vietnamese_VLC_v2_2022"
        DataFetcher.download_data(data, None)



# flake8: noqa

from argparse import Namespace
from pathlib import Path
from typing import NamedTuple

from jax.numpy import ndarray


class FLAGS(Namespace):
    """Configurations"""

    duration_lstm_dim = 256
    vocab_size = 256
    duration_embed_dropout_rate = 0.5
    num_training_steps = 200_000
    postnet_dim = 512
    acoustic_decoder_dim = 512
    acoustic_encoder_dim = 256

    # dataset
    max_phoneme_seq_len = 256 * 1
    assert max_phoneme_seq_len % 256 == 0  # prevent compilation error on Colab T4 GPU
    max_wave_len = 1024 * 64 * 3

    # Montreal Forced Aligner
    special_phonemes = ["sil", "sp", "spn", " "]  # [sil], [sp] [spn] [word end]
    sil_index = special_phonemes.index("sil")
    sp_index = sil_index  # no use of "sp"
    word_end_index = special_phonemes.index(" ")
    _normal_phonemes = (
        [] +
        ["a", "b", "c", "d", "e", "g", "h", "i", "k", "l"] +
        ["m", "n", "o", "p", "q", "r", "s", "t", "u", "v"] +
        ["x", "y", "à", "á", "â", "ã", "è", "é", "ê", "ì"] +
        ["í", "ò", "ó", "ô", "õ", "ù", "ú", "ý", "ă", "đ"] +
        ["ĩ", "ũ", "ơ", "ư", "ạ", "ả", "ấ", "ầ", "ẩ", "ẫ"] +
        ["ậ", "ắ", "ằ", "ẳ", "ẵ", "ặ", "ẹ", "ẻ", "ẽ", "ế"] +
        ["ề", "ể", "ễ", "ệ", "ỉ", "ị", "ọ", "ỏ", "ố", "ồ"] +
        ["ổ", "ỗ", "ộ", "ớ", "ờ", "ở", "ỡ", "ợ", "ụ", "ủ"] +
        ["ứ", "ừ", "ử", "ữ", "ự", "ỳ", "ỵ", "ỷ", "ỹ"]
    )

    # dsp
    mel_dim = 80
    n_fft = 1024
    sample_rate = 16000
    fmin = 0.0
    fmax = 8000

    # training
    batch_size = 64
    learning_rate = 1e-4
    duration_learning_rate = 1e-4
    max_grad_norm = 1.0
    weight_decay = 1e-4
    token_mask_prob = 0.1

    # ckpt
    ckpt_dir = Path("assets/infore/nat")
    data_dir = Path("train_data")


class DurationInput(NamedTuple):
    phonemes: ndarray
    lengths: ndarray
    durations: ndarray


class AcousticInput(NamedTuple):
    phonemes: ndarray
    lengths: ndarray
    durations: ndarray
    wavs: ndarray
    wav_lengths: ndarray
    mels: ndarray



from os.path import join, dirname

import joblib


class TextNormalizer:
    def __init__(self, binary_file):
        data = joblib.load(binary_file)
        self.character_map = data["character"]
        self.token_map = data["token"]


# rule_file = "tn_rules_2022_08_11.bin"
rule_file = "tn_rules_2023_07_14.bin"
text_normalizer = TextNormalizer(join(dirname(__file__), rule_file))
character_map = text_normalizer.character_map
token_map = text_normalizer.token_map
text_normalizer = TextNormalizer(join(dirname(__file__), rule_file))
character_map = text_normalizer.character_map
token_map = text_normalizer.token_map



import haiku as hk
import jax
import jax.numpy as jnp

LRELU_SLOPE = 0.1


def get_padding(kernel_size, dilation=1):
    p = int((kernel_size * dilation - dilation) / 2)
    return ((p, p),)


class ResBlock1(hk.Module):
    def __init__(
        self, h, channels, kernel_size=3, dilation=(1, 3, 5), name="resblock1"
    ):
        super().__init__(name=name)

        self.h = h
        self.convs1 = [
            hk.Conv1D(
                channels,
                kernel_size,
                1,
                rate=dilation[i],
                padding=get_padding(kernel_size, dilation[i]),
                name=f"convs1_{i}",
            )
            for i in range(3)
        ]

        self.convs2 = [
            hk.Conv1D(
                channels,
                kernel_size,
                1,
                rate=1,
                padding=get_padding(kernel_size, 1),
                name=f"convs2_{i}",
            )
            for i in range(3)
        ]

    def __call__(self, x):
        for c1, c2 in zip(self.convs1, self.convs2):
            xt = jax.nn.leaky_relu(x, LRELU_SLOPE)
            xt = c1(xt)
            xt = jax.nn.leaky_relu(xt, LRELU_SLOPE)
            xt = c2(xt)
            x = xt + x
        return x


class ResBlock2(hk.Module):
    def __init__(self, h, channels, kernel_size=3, dilation=(1, 3), name="ResBlock2"):
        super().__init__(name=name)
        self.h = h
        self.convs = [
            hk.Conv1D(
                channels,
                kernel_size,
                1,
                rate=dilation[i],
                padding=get_padding(kernel_size, dilation[i]),
            )
            for i in range(2)
        ]

    def __call__(self, x):
        for c in self.convs:
            xt = jax.nn.leaky_relu(x, LRELU_SLOPE)
            xt = c(xt)
            x = xt + x
        return x


class Generator(hk.Module):
    def __init__(self, h):
        super().__init__()
        self.h = h
        self.num_kernels = len(h.resblock_kernel_sizes)
        self.num_upsamples = len(h.upsample_rates)
        self.conv_pre = hk.Conv1D(h.upsample_initial_channel, 7, 1, padding=((3, 3),))
        resblock = ResBlock1 if h.resblock == "1" else ResBlock2
        self.ups = []
        for i, (u, k) in enumerate(zip(h.upsample_rates, h.upsample_kernel_sizes)):
            self.ups.append(
                hk.Conv1DTranspose(
                    h.upsample_initial_channel // (2 ** (i + 1)),
                    kernel_shape=k,
                    stride=u,
                    padding="SAME",
                    name=f"ups_{i}",
                )
            )

        self.resblocks = []

        for i in range(len(self.ups)):
            ch = h.upsample_initial_channel // (2 ** (i + 1))
            for j, (k, d) in enumerate(
                zip(h.resblock_kernel_sizes, h.resblock_dilation_sizes)
            ):
                self.resblocks.append(
                    resblock(h, ch, k, d, name=f"res_block1_{len(self.resblocks)}")
                )
        self.conv_post = hk.Conv1D(1, 7, 1, padding=((3, 3),))

    def __call__(self, x):
        x = self.conv_pre(x)
        for i in range(self.num_upsamples):
            x = jax.nn.leaky_relu(x, LRELU_SLOPE)

            x = self.ups[i](x)
            xs = None
            for j in range(self.num_kernels):
                if xs is None:
                    xs = self.resblocks[i * self.num_kernels + j](x)
                else:
                    xs += self.resblocks[i * self.num_kernels + j](x)
            x = xs / self.num_kernels
        x = jax.nn.leaky_relu(x)  # default pytorch value
        x = self.conv_post(x)
        x = jnp.tanh(x)
        return x



from os.path import join
from pathlib import Path
import joblib
import pycrfsuite
from underthesea_core import CRFFeaturizer


class FastCRFSequenceTagger:
    def __init__(self, features=None, dictionary=set()):
        self.features = features
        self.dictionary = dictionary
        self.estimator = None
        self.featurizer = None
        self.path_model = "models.bin"
        self.path_features = "features.bin"
        self.path_dictionary = "dictionary.bin"

    def forward(self, samples, contains_labels=False):
        return self.featurizer.process(samples, contains_labels)

    def save(self, base_path):
        joblib.dump(self.features, join(base_path, self.path_features))
        joblib.dump(self.dictionary, join(base_path, self.path_dictionary))

    def load(self, base_path):
        # print(base_path)
        model_path = str(Path(base_path) / self.path_model)
        estimator = pycrfsuite.Tagger()
        estimator.open(model_path)
        features = joblib.load(join(base_path, self.path_features))
        dictionary = joblib.load(join(base_path, self.path_dictionary))
        featurizer = CRFFeaturizer(features, dictionary)
        self.featurizer = featurizer
        self.estimator = estimator

    def predict(self, tokens):
        x = self.featurizer.process([tokens])[0]
        tags = self.estimator.tag(x)
        return tags



# Python code for youtube.playlistItems.list
# To set up YouTube API: https://developers.google.com/youtube/v3/getting-started
# To use channel PlaylistItems: https://developers.google.com/youtube/v3/docs/playlistItems/list

import os
import google_auth_oauthlib.flow
import googleapiclient.discovery
import googleapiclient.errors
from os.path import dirname, join
from datetime import datetime
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import TextFormatter

PROJECT_FOLDER = dirname(dirname(dirname(__file__)))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")

scopes = ["https://www.googleapis.com/auth/youtube.readonly"]


def playlist_youtube_api(playlist):
    # Disable OAuthlib's HTTPS verification when running locally.
    # *DO NOT* leave this option enabled in production.
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

    api_service_name = "youtube"
    api_version = "v3"
    client_secrets_file = "client_secret.json"

    # Get credentials and create an API client
    flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(
        client_secrets_file, scopes)
    credentials = flow.run_console()
    youtube = googleapiclient.discovery.build(
        api_service_name, api_version, credentials=credentials)

    request = youtube.playlistItems().list(
        part="contentDetails",
        maxResults=50,
        playlistId=playlist
    )
    response = request.execute()
    return response


class COLReader:
    def __init__(self, response):
        self.response = response
        self.items = self.response["items"]

        data = []
        for item in self.items:
            id = item["contentDetails"]["videoId"]
            doc_url = "".join(["https://www.youtube.com/watch?v=", item["contentDetails"]["videoId"]])
            date = datetime.strptime(item["contentDetails"]["videoPublishedAt"], "%Y-%m-%dT%H:%M:%SZ").strftime("%Y%m%d")
            try:
                transcript = TextFormatter().format_transcript(
                    YouTubeTranscriptApi.get_transcript(id, languages=["vi"]))
            except Exception:
                print("Could not extract video: %s" % doc_url)

            if transcript:
                sentences = transcript.split("\n")
                for sentence in sentences:
                    if len(sentence.replace("♪", "").strip()):
                        s = {
                            "doc_url": doc_url,
                            "date": date,
                            "sentence": sentence.replace("♪", "").strip()
                        }

                        data.append(s)

        self.data = data

    def __len__(self):
        return len(self.items)


if __name__ == "__main__":
    playlists = ["PLH_v4r_pvudV5ZrNx9HldKLICIjUSCRLb"]
    print("YouTube playlists: %s" % len(playlists))
    print("You will be required to authenticate each!")

    # Get all lyrics from playlists
    all_data = []
    for pl in playlists:
        response = playlist_youtube_api(pl)
        data = COLReader(response).data
        print("Playlist videos: %s" % len(data))
        all_data.extend(data)
    print("Sample\n", all_data[0])
    print("Total lyric lines %s" % len(all_data))

    # Write to file
    content = ""
    for i, s in enumerate(all_data):
        doc_url = "# doc_url = " + s["doc_url"]
        date = "# date = " + s["date"]
        sent_id = "# sent_id = " + str(i + 1)
        sent = s["sentence"]
        content += "\n".join([doc_url, date, sent_id, sent, "\n"])

    target_file = join(COL_FOLDER, "corpus", "raw", "lyrics.txt")
    with open(target_file, "w") as f:
        f.write(content)



from os.path import join

from underthesea.corpus.tagged_corpus import TaggedCorpus


class WordTokenizeCorpusReader:
    @staticmethod
    def read(data_folder, train_file=None, test_file=None):
        train = WordTokenizeCorpusReader.__read_data(join(data_folder, train_file))
        test = WordTokenizeCorpusReader.__read_data(join(data_folder, test_file))
        tagged_corpus = TaggedCorpus(train, test)
        return tagged_corpus

    @staticmethod
    def __read_data(data_file):
        text = open(data_file).read()
        sentences = text.split("\n")
        sentences = [WordTokenizeCorpusReader.__extract_tokens(s) for s in sentences]
        return sentences

    @staticmethod
    def __extract_tokens(s):
        sentence = []
        for item in s.split():
            tokens = item.split("_")
            tokens = [token for token in tokens if token]
            for i, token in enumerate(tokens):
                if i == 0:
                    sentence.append((token, "B-W"))
                else:
                    sentence.append((token, "I-W"))
        return sentence



# -*- coding: utf-8 -*-
from collections import Counter
import torch
from underthesea.utils.sp_fn import pad
from underthesea.utils.sp_vocab import Vocab


class RawField(object):
    r"""
    Defines a general datatype.

    A :class:`RawField` object does not assume any property of the datatype and
    it holds parameters relating to how a datatype should be processed.

    Args:
        name (str):
            The name of the field.
        fn (function):
            The function used for preprocessing the examples. Default: ``None``.
    """

    def __init__(self, name, fn=None):
        self.name = name
        self.fn = fn

    def __repr__(self):
        return f"({self.name}): {self.__class__.__name__}()"

    def preprocess(self, sequence):
        return self.fn(sequence) if self.fn is not None else sequence

    def transform(self, sequences):
        return [self.preprocess(seq) for seq in sequences]

    def compose(self, sequences):
        return sequences


class Field(RawField):
    r"""
    Defines a datatype together with instructions for converting to :class:`~torch.Tensor`.
    :class:`Field` models common text processing datatypes that can be represented by tensors.
    It holds a :class:`Vocab` object that defines the set of possible values
    for elements of the field and their corresponding numerical representations.
    The :class:`Field` object also holds other parameters relating to how a datatype
    should be numericalized, such as a tokenization method.

    Args:
        name (str):
            The name of the field.
        pad_token (str):
            The string token used as padding. Default: ``None``.
        unk_token (str):
            The string token used to represent OOV words. Default: ``None``.
        bos_token (str):
            A token that will be prepended to every example using this field, or ``None`` for no `bos_token`.
            Default: ``None``.
        eos_token (str):
            A token that will be appended to every example using this field, or ``None`` for no `eos_token`.
        lower (bool):
            Whether to lowercase the text in this field. Default: ``False``.
        use_vocab (bool):
            Whether to use a :class:`Vocab` object. If ``False``, the data in this field should already be numerical.
            Default: ``True``.
        tokenize (function):
            The function used to tokenize strings using this field into sequential examples. Default: ``None``.
        fn (function):
            The function used for preprocessing the examples. Default: ``None``.
    """

    def __init__(self, name, pad=None, unk=None, bos=None, eos=None,
                 lower=False, use_vocab=True, tokenize=None, fn=None):
        self.name = name
        self.pad = pad
        self.unk = unk
        self.bos = bos
        self.eos = eos
        self.lower = lower
        self.use_vocab = use_vocab
        self.tokenize = tokenize
        self.fn = fn

        self.specials = [token for token in [pad, unk, bos, eos]
                         if token is not None]

    def __repr__(self):
        s, params = f"({self.name}): {self.__class__.__name__}(", []
        if self.pad is not None:
            params.append(f"pad={self.pad}")
        if self.unk is not None:
            params.append(f"unk={self.unk}")
        if self.bos is not None:
            params.append(f"bos={self.bos}")
        if self.eos is not None:
            params.append(f"eos={self.eos}")
        if self.lower:
            params.append(f"lower={self.lower}")
        if not self.use_vocab:
            params.append(f"use_vocab={self.use_vocab}")
        s += ", ".join(params)
        s += ")"

        return s

    @property
    def pad_index(self):
        if self.pad is None:
            return 0
        if hasattr(self, 'vocab'):
            return self.vocab[self.pad]
        return self.specials.index(self.pad)

    @property
    def unk_index(self):
        if self.unk is None:
            return 0
        if hasattr(self, 'vocab'):
            return self.vocab[self.unk]
        return self.specials.index(self.unk)

    @property
    def bos_index(self):
        if hasattr(self, 'vocab'):
            return self.vocab[self.bos]
        return self.specials.index(self.bos)

    @property
    def eos_index(self):
        if hasattr(self, 'vocab'):
            return self.vocab[self.eos]
        return self.specials.index(self.eos)

    @property
    def device(self):
        return 'cuda' if torch.cuda.is_available() else 'cpu'

    def preprocess(self, sequence):
        r"""
        Loads a single example using this field, tokenizing if necessary.
        The sequence will be first passed to ``fn`` if available.
        If ``tokenize`` is not None, the input will be tokenized.
        Then the input will be lowercased optionally.

        Args:
            sequence (list):
                The sequence to be preprocessed.

        Returns:
            A list of preprocessed sequence.
        """

        if self.fn is not None:
            sequence = self.fn(sequence)
        if self.tokenize is not None:
            sequence = self.tokenize(sequence)
        if self.lower:
            sequence = [str.lower(token) for token in sequence]

        return sequence

    def vi_preprocess(self, sequence):
        sequence = [[token] for token in sequence]
        return sequence

    def build(self, dataset, min_freq=1, embed=None):
        r"""
        Constructs a :class:`Vocab` object for this field from the dataset.
        If the vocabulary has already existed, this function will have no effect.

        Args:
            dataset (Dataset):
                A :class:`Dataset` object. One of the attributes should be named after the name of this field.
            min_freq (int):
                The minimum frequency needed to include a token in the vocabulary. Default: 1.
            embed (Embedding):
                An Embedding object, words in which will be extended to the vocabulary. Default: ``None``.
        """

        if hasattr(self, 'vocab'):
            return
        sequences = getattr(dataset, self.name)
        counter = Counter(token
                          for seq in sequences
                          for token in self.preprocess(seq))
        self.vocab = Vocab(counter, min_freq, self.specials, self.unk_index)

        if not embed:
            self.embed = None
        else:
            tokens = self.preprocess(embed.tokens)
            # if the `unk` token has existed in the pretrained,
            # then replace it with a self-defined one
            if embed.unk:
                tokens[embed.unk_index] = self.unk

            self.vocab.extend(tokens)
            self.embed = torch.zeros(len(self.vocab), embed.dim)
            self.embed[self.vocab[tokens]] = embed.vectors
            self.embed /= torch.std(self.embed)

    def transform(self, sequences):
        r"""
        Turns a list of sequences that use this field into tensors.

        Each sequence is first preprocessed and then numericalized if needed.

        Args:
            sequences (list[list[str]]):
                A list of sequences.

        Returns:
            A list of tensors transformed from the input sequences.
        """

        sequences = [self.preprocess(seq) for seq in sequences]
        if self.use_vocab:
            sequences = [self.vocab[seq] for seq in sequences]
        if self.bos:
            sequences = [[self.bos_index] + seq for seq in sequences]
        if self.eos:
            sequences = [seq + [self.eos_index] for seq in sequences]
        sequences = [torch.tensor(seq) for seq in sequences]

        return sequences

    def compose(self, sequences):
        r"""
        Composes a batch of sequences into a padded tensor.

        Args:
            sequences (list[~torch.Tensor]):
                A list of tensors.

        Returns:
            A padded tensor converted to proper device.
        """

        return pad(sequences, self.pad_index).to(self.device)


class SubwordField(Field):
    r"""
    A field that conducts tokenization and numericalization over each token rather the sequence.

    This is customized for models requiring character/subword-level inputs, e.g., CharLSTM and BERT.

    Args:
        fix_len (int):
            A fixed length that all subword pieces will be padded to.
            This is used for truncating the subword pieces that exceed the length.
            To save the memory, the final length will be the smaller value
            between the max length of subword pieces in a batch and `fix_len`.

    Examples:
        >>> from transformers import AutoTokenizer
        >>> tokenizer = AutoTokenizer.from_pretrained('bert-base-cased')
        >>> field = SubwordField('bert',
                                 pad=tokenizer.pad_token,
                                 unk=tokenizer.unk_token,
                                 bos=tokenizer.cls_token,
                                 eos=tokenizer.sep_token,
                                 fix_len=20,
                                 tokenize=tokenizer.tokenize)
        >>> field.vocab = tokenizer.get_vocab()  # no need to re-build the vocab
        >>> field.transform([['This', 'field', 'performs', 'token-level', 'tokenization']])[0]
        tensor([[  101,     0,     0],
                [ 1188,     0,     0],
                [ 1768,     0,     0],
                [10383,     0,     0],
                [22559,   118,  1634],
                [22559,  2734,     0],
                [  102,     0,     0]])
    """

    def __init__(self, *args, **kwargs):
        self.fix_len = kwargs.pop('fix_len') if 'fix_len' in kwargs else 0
        super().__init__(*args, **kwargs)

    def build(self, dataset, min_freq=1, embed=None):
        if hasattr(self, 'vocab'):
            return
        sequences = getattr(dataset, self.name)
        counter = Counter(piece
                          for seq in sequences
                          for token in seq
                          for piece in self.preprocess(token))
        self.vocab = Vocab(counter, min_freq, self.specials, self.unk_index)

        if not embed:
            self.embed = None
        else:
            tokens = self.preprocess(embed.tokens)
            # if the `unk` token has existed in the pretrained,
            # then replace it with a self-defined one
            if embed.unk:
                tokens[embed.unk_index] = self.unk

            self.vocab.extend(tokens)
            self.embed = torch.zeros(len(self.vocab), embed.dim)
            self.embed[self.vocab[tokens]] = embed.vectors

    def transform(self, sequences):
        sequences_ = [self.vi_preprocess(seq) for seq in sequences]
        sequences_ = [self.vi_preprocess(seq) for seq in sequences]
        sequences = sequences_
        if self.fix_len <= 0:
            self.fix_len = max(len(token) for seq in sequences for token in seq)

        def get_vocab(i):
            if i in self.vocab:
                return self.vocab[i]
            else:
                return self.vocab['<unk>']

        if self.use_vocab:
            sequences = [[[get_vocab(i) for i in token] if token else [self.unk_index] for token in seq] for seq in
                         sequences]
        if self.bos:
            sequences = [[[self.bos_index]] + seq for seq in sequences]
        if self.eos:
            sequences = [seq + [[self.eos_index]] for seq in sequences]
        lens = [min(self.fix_len, max(len(ids) for ids in seq)) for seq in sequences]
        sequences = [pad([torch.tensor(ids[:i]) for ids in seq], self.pad_index, i)
                     for i, seq in zip(lens, sequences)]

        return sequences


class ChartField(Field):
    r"""
    Field dealing with constituency trees.

    This field receives sequences of binarized trees factorized in pre-order,
    and returns two tensors representing the bracketing trees and labels on each constituent respectively.

    Examples:
        >>> sequence = [(0, 5, 'S'), (0, 4, 'S|<>'), (0, 1, 'NP'), (1, 4, 'VP'), (1, 2, 'VP|<>'),
                        (2, 4, 'S+VP'), (2, 3, 'VP|<>'), (3, 4, 'NP'), (4, 5, 'S|<>')]
        >>> spans, labels = field.transform([sequence])[0]  # this example field is built from ptb
        >>> spans
        tensor([[False,  True, False, False,  True,  True],
                [False, False,  True, False,  True, False],
                [False, False, False,  True,  True, False],
                [False, False, False, False,  True, False],
                [False, False, False, False, False,  True],
                [False, False, False, False, False, False]])
        >>> labels
        tensor([[  0,  37,   0,   0, 107,  79],
                [  0,   0, 120,   0, 112,   0],
                [  0,   0,   0, 120,  86,   0],
                [  0,   0,   0,   0,  37,   0],
                [  0,   0,   0,   0,   0, 107],
                [  0,   0,   0,   0,   0,   0]])
    """

    def build(self, dataset, min_freq=1):
        counter = Counter(label
                          for seq in getattr(dataset, self.name)
                          for i, j, label in self.preprocess(seq))

        self.vocab = Vocab(counter, min_freq, self.specials, self.unk_index)

    def transform(self, sequences):
        sequences = [self.preprocess(seq) for seq in sequences]
        spans, labels = [], []

        for sequence in sequences:
            seq_len = sequence[0][1] + 1
            span_chart = torch.full((seq_len, seq_len), self.pad_index, dtype=torch.bool)
            label_chart = torch.full((seq_len, seq_len), self.pad_index, dtype=torch.long)
            for i, j, label in sequence:
                span_chart[i, j] = 1
                label_chart[i, j] = self.vocab[label]
            spans.append(span_chart)
            labels.append(label_chart)

        return list(zip(spans, labels))

    def compose(self, sequences):
        return [pad(i).to(self.device) for i in zip(*sequences)]



import unicodedata

from .text_normalizer import character_map


def character_normalize(text):
    for character_non_standard in character_map:
        character_standard = character_map[character_non_standard]
        text = text.replace(character_non_standard, character_standard)
    return text


def utf8_normalize(text):
    if not is_unicode(text):
        text = text.decode("utf-8")
    text = unicodedata.normalize("NFC", text)
    return text


def normalize_characters_in_text(text):
    text = utf8_normalize(text)
    text = character_normalize(text)
    return text


def is_unicode(text):
    return type(text) is str



from os.path import join, dirname
from underthesea.reader.dictionary_loader import DictionaryLoader

words = DictionaryLoader(join(dirname(__file__), "Viet74K.txt")).words
lower_words = set([word.lower() for word in words])


def text_lower(word):
    return word.lower()


def text_isdigit(word):
    return word.isdigit()


def text_isallcap(word):
    for letter in word:
        if not letter.istitle():
            return False
    return True


def text_istitle(word):
    if len(word) == 0:
        return False
    try:
        titles = [s[0] for s in word.split(" ")]
        for token in titles:
            if token[0].istitle() is False:
                return False
        return True
    except Exception:
        return False


def text_is_in_dict(word):
    return str(word.lower() in lower_words)


functions = {
    "lower": text_lower,
    "istitle": text_istitle,
    "isallcap": text_isallcap,
    "isdigit": text_isdigit,
    "is_in_dict": text_is_in_dict
}



import torch.nn


class WordEmbeddings(torch.nn.Module):
    def __init__(
        self,
        n_word: int = 100,
        learn=False
    ):
        if learn:
            return
        self.n_word = n_word

    @classmethod
    def fit(cls, sentences):
        n_word = 100
        embed = cls(
            n_word=n_word
        )
        return embed


class CharacterEmbeddings(torch.nn.Module):
    def __init__(self):
        pass

    def fit(self, sentences):
        pass


class FieldEmbeddings:
    def __init__(self):
        self._n_vocab = None

    @property
    def n_vocab(self) -> int:
        return self._n_vocab

    @n_vocab.setter
    def n_vocab(self, val):
        self._n_vocab = val

    def __repr__(self):
        s = f"{self.__class__.__name__}(n_vocab={self._n_vocab})"
        return s


class TokenEmbeddings(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.n_words = None
        self.pad_index = None

    def build(self, sentences):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}({self.n_words})"


class StackEmbeddings(torch.nn.Module):
    def __init__(self, embeddings=[]):
        super().__init__()
        self.embeddings = embeddings

    def __repr__(self):
        return f"{self.__class__.__name__}()"
